Package: netUtils
Title: A Collection of Tools for Network Analysis
Version: 0.8.6
Authors@R: 
    person("David", "Schoch", email = "david@schochastics.net", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2952-4812"))
Description: Provides a collection of network analytic (convenience) functions which are missing in other standard packages. This includes triad census with attributes <doi:10.1016/j.socnet.2019.04.003>, core-periphery models <doi:10.1016/S0378-8733(99)00019-2>, and several graph generators. Most functions are build upon 'igraph'.  
URL: https://github.com/schochastics/netUtils/,
        https://schochastics.github.io/netUtils/
BugReports: https://github.com/schochastics/netUtils/issues
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
LinkingTo: Rcpp, RcppArmadillo
Imports: Rcpp, igraph (>= 2.0.0), stats
Suggests: covr, GA, testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-06-12 19:23:15 UTC; david
Author: David Schoch [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2952-4812>)
Maintainer: David Schoch <david@schochastics.net>
Repository: CRAN
Date/Publication: 2026-06-12 20:40:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-06-12 21:27:25 UTC; unix
Archs: netUtils.so.dSYM

Package: netUtils
Title: A Collection of Tools for Network Analysis
Version: 0.8.6
Authors@R: 
    person("David", "Schoch", email = "david@schochastics.net", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2952-4812"))
Description: Provides a collection of network analytic (convenience) functions which are missing in other standard packages. This includes triad census with attributes <doi:10.1016/j.socnet.2019.04.003>, core-periphery models <doi:10.1016/S0378-8733(99)00019-2>, and several graph generators. Most functions are build upon 'igraph'.  
URL: https://github.com/schochastics/netUtils/,
        https://schochastics.github.io/netUtils/
BugReports: https://github.com/schochastics/netUtils/issues
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
LinkingTo: Rcpp, RcppArmadillo
Imports: Rcpp, igraph (>= 2.0.0), stats
Suggests: covr, GA, testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-06-12 19:23:15 UTC; david
Author: David Schoch [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2952-4812>)
Maintainer: David Schoch <david@schochastics.net>
Repository: CRAN
Date/Publication: 2026-06-12 20:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(graph_cor,array)
S3method(graph_cor,default)
S3method(graph_cor,igraph)
S3method(graph_cor,matrix)
S3method(str,igraph)
export(as_adj_list1)
export(as_adj_weighted)
export(as_multi_adj)
export(biggest_component)
export(bipartite_from_data_frame)
export(clique_vertex_mat)
export(core_periphery)
export(delete_isolates)
export(dyad_census_attr)
export(graph_cartesian)
export(graph_cor)
export(graph_direct)
export(graph_from_multi_edgelist)
export(graph_kpartite)
export(reciprocity_cor)
export(sample_coreseq)
export(sample_lfr)
export(sample_pa_homophilic)
export(split_graph)
export(structural_equivalence)
export(triad_census_attr)
importFrom(Rcpp,sourceCpp)
useDynLib(netUtils, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# netUtils <img src="man/figures/logo.png" align="right"/>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/netUtils)](https://CRAN.R-project.org/package=netUtils)
[![CRAN
Downloads](https://cranlogs.r-pkg.org/badges/netUtils)](https://CRAN.R-project.org/package=netUtils)
[![R-CMD-check](https://github.com/schochastics/netUtils/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/schochastics/netUtils/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/schochastics/netUtils/graph/badge.svg)](https://app.codecov.io/gh/schochastics/netUtils)
<!-- badges: end -->

netUtils is a collection of tools for network analysis that may not
deserve a package on their own and/or are missing from other network
packages.

## Installation

You can install the development version of netUtils with:

``` r
# install.packages("remotes")
remotes::install_github("schochastics/netUtils")
```

## Functions

most functions only support igraph objects

**helper/convenience functions**  
`biggest_component()` extracts the biggest connected component of a
network.  
`delete_isolates()` deletes vertices with degree zero.  
`bipartite_from_data_frame()` creates a two mode network from a data
frame.  
`graph_from_multi_edgelist()` creates multiple graphs from a typed
edgelist.  
`clique_vertex_mat()` computes the clique vertex matrix.  
`graph_cartesian()` computes the Cartesian product of two graphs.  
`graph_direct()` computes the direct (or tensor) product of graphs.  
`str()` extends str to work with igraph objects.

**methods**  
`dyad_census_attr()` calculates dyad census with node attributes.  
`triad_census_attr()` calculates triad census with node attributes.  
`core_periphery()` fits a discrete core periphery model.  
`graph_kpartite()` creates a random k-partite network.  
`split_graph()` sample graph with perfect core periphery structure.  
`sample_coreseq()` creates a random graph with given coreness
sequence.  
`sample_pa_homophilic()` creates a preferential attachment graph with
two groups of nodes.  
`sample_lfr()` create LFR benchmark graph for community detection.  
`structural_equivalence()` finds structurally equivalent vertices.  
`reciprocity_cor()` reciprocity as a correlation coefficient.

**methods to use with caution**  
*(this functions should only be used if you know what you are doing)*  
`as_adj_list1()` extracts the adjacency list faster, but less stable,
from igraph objects.  
`as_adj_weighted()` extracts the dense weighted adjacency matrix fast.

# netutils

A Python library that is a collection of objects for common network automation tasks.

This library intends to keep the following tenets:

* Must not be any dependencies required to run the library.
  * May be some optional dependencies, to be managed by the user in opt in fashion.
* Shall prefer functions over classes.
* Shall prefer a folder and file structure that is flat.
* Shall leverage docstrings as the primary documentation mechanism.
  * Must provide examples in every public function.
* Shall retain a high test coverage.

# Function Groupings

Functions are grouped with like functions, such as IP or MAC address based functions. Included to date are groupings of:

* BGP ASN - Provides the ability to convert BGP ASN from integer to dot notation.
* Configuration
  * Cleaning - Provides the ability to remove or replace lines based on regex matches.
  * Compliance - Provides the ability to compare two configurations to sanely understand the differences.
  * Parsing - Provides the ability to parse configuration for the minor differences that are there.
* DNS - Provides the ability to work with DNS, such as validating that a FQDN is resolvable.
* Interface - Provides the ability to work with interface names, expanding, abbreviating, and spliting the names.
* IP Address - Provides the ability to work with IP addresses, primarily exposing Python `ipaddress` functionality.
* Library Mapper - Provides mappings in expected vendor names between Netmiko, NAPALM, pyntc, ntc-templates, pyats, and scrapli.
* MAC Address - Provides the ability to work with MAC addresses such as validating or converting to integer.
* Password - Provides the ability to compare and encrypt common password schemas such as type5 and type7 Cisco passwords.
* Ping - Provides the ability to ping, currently only tcp ping.
* Protocol Mapper - Provides a mapping for protocol names to numbers and vice versa.
* Route - Provides the ability to provide a list of routes and an IP Address and return the longest prefix matched route.
* VLANs - Provide the ability to convert configuration into lists or lists into configuration.

# Installation

Option 1: Install from PyPI.

```bash
$ pip install netutils
```

Option 2: Install from a GitHub branch, such as develop as shown below.

```bash
$ pip install git+https://github.com/networktocode/netutils.git@develop
```

# Examples

While all functions come with examples in the docstrings, for quick reference of the types of problems this library intends to
solve the following examples are provided.

The following function will help in deploying list of VLANs and match the configuration style in a standard IOS-like configurations.

```python
>>> from netutils.vlan import vlanlist_to_config
>>>
>>> vlan_cfg = vlanlist_to_config([1, 2, 3, 5, 6, 1000, 1002, 1004, 1006, 1008, 1010, 1012, 1014, 1016, 1018])
>>>
>>> vlan_cfg
["1-3,5,6,1000,1002,1004,1006,1008,1010,1012,1014", "1016,1018"]
>>>
>>> for index, line in enumerate(vlan_cfg):
...     if index == 0:
...         print(f"  switchport trunk allowed vlan {line}")
...     else:
...         print(f"  switchport trunk allowed vlan add {line}")
...
  switchport trunk allowed vlan 1-3,5,6,1000,1002,1004,1006,1008,1010,1012,1014
  switchport trunk allowed vlan add 1016,1018
>>>
```

You may want to compare a known password with a given encrypted password. This can help in verifying if the
passwords are as expected for compliance reasons.

```python
>>> from netutils.password import compare_type5
>>>
>>> compare_type5("cisco","$1$nTc1$Z28sUTcWfXlvVe2x.3XAa.")
True
>>>
>>> compare_type5("not_cisco","$1$nTc1$Z28sUTcWfXlvVe2x.3XAa.")
False
>>>
```

Often times interfaces will come in various different shortened names, and it is helpful to normalize them.

```python
>>> from netutils.interface import canonical_interface_name
>>>
>>> canonical_interface_name("Gi1/0/1")
'GigabitEthernet1/0/1'
>>>
>>> canonical_interface_name("Eth1")
'Ethernet1'
>>>
```

These are just some examples of the many functions provided by this library.

# Attribution

The library was built to be a centralized place for common network automation code to be accessed. While in most cases it is
difficult, if not impossible to understand the origin of code, the following intends to describe the known motivation for where
code was derived from and in the few cases where actual code was directly taken from. Except where noted, all code is believed to
be unattributable, from @itdependsnetworks, or from another Network to Code employee. If this is in fact an error, please open an
issue, and the proper attribution will be provided. Any errors were not done out of malice, but rather the natural developer
workflow of pulling snippets of code from existing locations such as StackOverflow and Github over months and years of development.
As an example it is nearly impossible to understand the original author of Cisco type7 encryption/decryption in the sea of
available code, and remains unattributable though clearly originally developed prior to this library being created.

Influencers
* [Netmiko](https://github.com/ktbyers/netmiko)
* [NAPALM](https://github.com/napalm-automation/napalm)
* [Ansible](https://github.com/ansible/ansible)
* [IPCal](https://github.com/ammyblabla/ipcal)
* [StackOverflow](https://stackoverflow.com/)
* [Python 3 Docs](https://docs.python.org/3/library/)

In many instances variables and function names were reused, but the code was built from scratch to avoid any potential licensing
issues. Functions that were known to be rewritten and their known origin.

| Function | Origin |
| -------- | ------ |
| asn_to_int | NAPALM |
| is_ip | IPCal |
| ip_to_bin | IPCal |
| get_usable_range | IPCal |
| encrypt_type7 | unknown |
| decrypt_type7 | unknown |
| vlan_to_list | Ansible |
| sanitize_config | NAPALM |

Relevant PR's
* https://github.com/napalm-automation/napalm/pull/493
* https://github.com/ansible/ansible/pull/39901
* https://github.com/ansible/ansible/pull/26566


# Contributing

Pull requests are welcomed and automatically built and tested against multiple versions of Python through TravisCI.
Except for unit tests, testing is only supported on Python 3.7.

The project is packaged with a light development environment based on `docker-compose` to help with the local development of the project and to run tests within TravisCI.

The project is following Network to Code software development guidelines and are leveraging the following:
- Black, Pylint, Bandit, flake8, and pydocstyle for Python linting and formatting.
- pytest, coverage, and unittest for unit tests.

There are a number of things that are required in order to have a successful PR.

- All new functions must contain at least 1 example in their docstrings.
- Docstrings must conform to the google docstring [convention](https://google.github.io/styleguide/pyguide.html#381-docstrings).
- Unit test for newly added functions are required.
- If applicable, tests related to config parsing and compliuance must be added.
- Update the jinja2 filter (netutils.utils.jinja2_convenience_function) for any new functions (see below for details).
- If you create a new file in the `netutils` folder, you must create a new folder and `index.rst` in the docs folder (see below for details).
- Your PR must not introduce any required dependencies. You can introduce optional or development dependencies.

## Adding to the jinja2 filter function

To add a new function to the jinja2 filter, add a new entry to the `_JINJA2_FUNCTION_MAPPINGS` located in the `utils.py` file. When adding an entry, the key corresponds with the name to call the function and the value to the path to find the function.

## Adding docs for a new python file

If adding a new python file, the docs must be updated to account for the new file.

1. Create a new folder in `docs/source/netutils` matching the name of your new file.
2. Create an `index.rst` file in that folder.
3. Add the following to the newly created file.

```python
#############################
# ENTER THE TITLE OF THE PAGE
##############################

.. automodule:: netutils.newfile
    :members:
```

## CLI Helper Commands

The project features a CLI helper based on [invoke](http://www.pyinvoke.org/) to help setup the development environment. The commands are listed below in 3 categories:
- `dev environment`
- `utility`
- `testing`

Each command can be executed with `invoke <command>`. Each command also has its own help `invoke <command> --help`

### Local dev environment

```
  build              Build all docker images.
  clean              Remove the project specific image.
  rebuild            Clean the Docker image and then rebuild without using cache.
```

### Utility

```
  clean-docs         Removes the build directory and all of its contents.
  check-pypi-version Verify if the version specified already exists on PyPI.
  cli                Enter the image to perform troubleshooting or dev work.
  html               Creates html docs using sphinx-build command.
```

### Testing

```
  bandit             Run bandit to validate basic static code security analysis.
  black              Run black to check that Python files adhere to its style standards.
  coverage           Run the coverage report against pytest.
  flake8             Run flake8 to check that Python files adhere to its style standards.
  pylint             Run pylint code analysis.
  pydocstyle         Run pydocstyle to validate docstring formatting adheres to NTC defined standards.
  pytest             Run pytest for the specified name and Python version.
  tests              Run all tests for the specified name and Python version.
  yamllint           Run yamllint to validate formatting adheres to NTC defined YAML standards.
```

## Questions

Please see [the documentation](https://netutils.readthedocs.io/) for detailed documentation on how to use netutils. For any additional questions or
comments, feel free to swing by the [Network to Code slack channel](https://networktocode.slack.com/) (channel #networktocode).
Sign up [here](http://slack.networktocode.com/)

#' LFR benchmark graphs
#' @description Generates benchmark networks for clustering tasks with a priori known communities. The algorithm accounts for the heterogeneity in the distributions of node degrees and of community sizes.
#' @param n Number of nodes in the created graph.
#' @param tau1 Power law exponent for the degree distribution of the created graph. This value must be strictly greater than one
#' @param tau2 Power law exponent for the community size distribution in the created graph. This value must be strictly greater than one
#' @param mu Fraction of inter-community edges incident to each node. This value must be in the interval 0 to 1.
#' @param average_degree Desired average degree of nodes in the created graph. This value must be in the interval 0 to n. Exactly one of this and `min_degree` must be specified, otherwise an error is raised
#' @param max_degree Maximum degree of nodes in the created graph. If not specified, this is set to n-1.
#' @param min_community Minimum size of communities in the graph. If not specified, this is set to `min_degree`
#' @param max_community Maximum size of communities in the graph. If not specified, this is set to n, the total number of nodes in the graph.
#' @param on number of overlapping nodes
#' @param om number of memberships of the overlapping nodes
#' @return an igraph object
#' @references A. Lancichinetti, S. Fortunato, and F. Radicchi.(2008) Benchmark graphs for testing community detection algorithms. Physical Review E, 78. arXiv:0805.4770
#' @details code adapted from <https://github.com/synwalk/synwalk-analysis/tree/master/lfr_generator>
#' @examples
#' # Simple Girven-Newman benchmark graphs
#' g <- sample_lfr(
#'     n = 128, average_degree = 16,
#'     max_degree = 16, mu = 0.1,
#'     min_community = 32, max_community = 32
#' )
#' @export
sample_lfr <- function(n,
                       tau1 = 2,
                       tau2 = 1,
                       mu = 0.1,
                       average_degree,
                       max_degree,
                       min_community = NULL,
                       max_community = NULL,
                       on = 0,
                       om = 0) {
    if (missing(average_degree)) {
        stop("average_degree must be specified")
    }

    if (average_degree > n || average_degree <= 0) {
        stop("average_degree must be in the interval (0, n]")
    }

    if (missing(max_degree)) {
        stop("max_degree must be specified")
    }
    if (max_degree > n || max_degree <= 0) {
        stop("max_degree must be in the interval (0, n]")
    }

    if (!(tau1 >= 1)) {
        stop("tau1 must be greater than one")
    }
    if (!(tau2 >= 1)) {
        stop("tau2 must be greater than one")
    }
    if (mu < 0 || mu > 1) {
        stop("mu must be in the interval [0, 1]")
    }

    if (!is.null(min_community) && !is.null(max_community)) {
        crange <- TRUE
    }

    if (is.null(min_community) && is.null(max_community)) {
        min_community <- 0
        max_community <- 0
        crange <- FALSE
    } else if (xor(is.null(min_community), is.null(max_community))) {
        stop("both min_community and max_community need to either be specified or not")
    }

    res <- benchmark(FALSE, FALSE,
        num_nodes = n,
        average_k = average_degree,
        max_degree = max_degree,
        tau = tau1,
        tau2 = tau2,
        mixing_parameter = mu,
        overlapping_nodes = on,
        overlap_membership = om,
        nmin = min_community,
        nmax = max_community,
        fixed_range = crange
    )

    g <- igraph::graph_from_adj_list(lapply(res[["edgelist"]], function(x) x + 1), mode = "all")
    igraph::V(g)$membership <- unlist(res$membership) + 1
    g
}

library(igraph)
test_that("lfr errors work", {
    # tau1 error
    expect_error(sample_lfr(
        n = 128,
        tau1 = 0.5,
        average_degree = 16,
        max_degree = 16,
        mu = 0.1,
        min_community = 32,
        max_community = 32
    ))
    # tau2 error
    expect_error(sample_lfr(
        n = 128,
        tau1 = 2,
        tau2 = 0.5,
        average_degree = 16,
        max_degree = 16,
        mu = 0.1,
        min_community = 32,
        max_community = 32
    ))
    # avg deg error
    expect_error(sample_lfr(
        n = 128,
        tau1 = 2,
        tau2 = 1,
        average_degree = 200,
        max_degree = 16,
        mu = 0.1,
        min_community = 32,
        max_community = 32
    ))
    # avg deg missing error
    expect_error(sample_lfr(
        n = 128,
        tau1 = 2,
        tau2 = 1,
        max_degree = 16,
        mu = 0.1,
        min_community = 32,
        max_community = 32
    ))
    # max deg error
    expect_error(sample_lfr(
        n = 128,
        tau1 = 2,
        tau2 = 1,
        average_degree = 16,
        max_degree = 200,
        mu = 0.1,
        min_community = 32,
        max_community = 32
    ))
    # mu error
    expect_error(sample_lfr(
        n = 128,
        tau1 = 2,
        tau2 = 1,
        average_degree = 16,
        max_degree = 16,
        mu = -1,
        min_community = 32,
        max_community = 32
    ))
    # com error
    expect_error(sample_lfr(
        n = 128,
        tau1 = 2,
        tau2 = 1,
        average_degree = 16,
        max_degree = 16,
        mu = 0.1,
        max_community = 32
    ))
})

test_that("lfr works", {
    g <- sample_lfr(
        n = 128,
        tau1 = 2,
        tau2 = 1,
        average_degree = 16,
        max_degree = 16,
        mu = 0.1,
        min_community = 32,
        max_community = 32
    )
    expect_equal(max(degree(g)), 16)
    expect_equal(max(V(g)$membership), 4)
    expect_equal(sum(V(g)$membership == 1), 32)
})

#' @title Maximal Structural Equivalence
#' @description Calculates structural equivalence for an undirected graph
#' @param g An igraph object
#' @details Two nodes u and v are structurally equivalent if they have exactly the same neighbors. The equivalence classes produced with this function are either cliques or empty graphs.
#' @return vector of equivalence classes
#' @author David Schoch
#' @export
structural_equivalence <- function(g) {
    if (igraph::is_directed(g)) {
        stop("g must be undirected")
    }
    adj <- lapply(igraph::neighborhood(g, mindist = 1), function(x) x - 1)
    deg <- igraph::degree(g)
    P <- mse(adj, deg)
    MSE <- which((P + t(P)) == 2, arr.ind = TRUE)
    if (length(MSE) >= 1) {
        MSE <- t(apply(MSE, 1, sort))
        MSE <- MSE[!duplicated(MSE), ]
        g <- igraph::make_empty_graph()
        g <- igraph::add_vertices(g, nrow(P))
        g <- igraph::add_edges(g, c(t(MSE)))
        g <- igraph::as_undirected(g)
        MSE <- igraph::components(g, "weak")$membership
    } else {
        MSE <- seq_len(nrow(P))
    }
    return(MSE)
}

test_that("mse works", {
    library(igraph)
    g1 <- make_full_graph(5, directed = FALSE)
    g2 <- make_full_graph(5, directed = TRUE)
    expect_equal(structural_equivalence(g1), c(1, 1, 1, 1, 1))
    expect_error(structural_equivalence(g2))
})

#' @title Maximal Structural Equivalence
#' @description Calculates structural equivalence for an undirected graph
#' @param g An igraph object
#' @details Two nodes u and v are structurally equivalent if they have exactly the same neighbors. The equivalence classes produced with this function are either cliques or empty graphs.
#' @return vector of equivalence classes
#' @author David Schoch
#' @export
structural_equivalence <- function(g) {
    if (igraph::is_directed(g)) {
        stop("g must be undirected")
    }
    adj <- lapply(igraph::neighborhood(g, mindist = 1), function(x) x - 1)
    deg <- igraph::degree(g)
    P <- mse(adj, deg)
    MSE <- which((P + t(P)) == 2, arr.ind = TRUE)
    if (length(MSE) >= 1) {
        MSE <- t(apply(MSE, 1, sort))
        MSE <- MSE[!duplicated(MSE), ]
        g <- igraph::make_empty_graph()
        g <- igraph::add_vertices(g, nrow(P))
        g <- igraph::add_edges(g, c(t(MSE)))
        g <- igraph::as.undirected(g)
        MSE <- igraph::components(g, "weak")$membership
    } else {
        MSE <- seq_len(nrow(P))
    }
    return(MSE)
}

#' @title two-mode network from a data.frame
#' @description Create a two-mode network from a data.frame
#'
#' @param d data.frame
#' @param type1 column name of mode 1
#' @param type2 column name of mode 2
#' @param attr named list of edge attributes
#' @param weighted should a weighted graph be created if multiple edges occur
#' @return two mode network as igraph object
#' @author David Schoch
#' @examples
#' library(igraph)
#' edges <- data.frame(mode1 = 1:5, mode2 = letters[1:5])
#' bipartite_from_data_frame(edges, "mode1", "mode2")
#' @export

bipartite_from_data_frame <- function(d, type1, type2, attr = NULL, weighted = TRUE) {
    if (!type1 %in% names(d)) {
        stop(paste0("no column named ", type1, " found in data frame"))
    }
    if (!type2 %in% names(d)) {
        stop(paste0("no column named ", type2, " found in data frame"))
    }

    mode1 <- unique(d[[type1]])
    mode2 <- unique(d[[type2]])
    if (any(mode1 %in% mode2)) {
        stop("some nodes appear in both modes. Modes in a two-mode network must be distinct.")
    }
    el <- cbind(d[[type1]], d[[type2]])

    g <- igraph::graph.empty(directed = FALSE)
    g <- igraph::add_vertices(g, nv = length(mode1), attr = list(name = mode1, type = TRUE))
    g <- igraph::add_vertices(g, nv = length(mode2), attr = list(name = mode2, type = FALSE))
    if (!is.null(attr)) {
        g <- igraph::add_edges(g, c(t(el)), attr = attr)
    } else {
        g <- igraph::add_edges(g, c(t(el)))
    }
    if (igraph::any_multiple(g) && weighted) {
        igraph::E(g)$weight <- 1
        g <- igraph::simplify(g,
            remove.multiple = TRUE,
            remove.loops = TRUE,
            edge.attr.comb = "sum"
        )
    }
    g
}

#' @title Multiple networks from a single edgelist with a typed attribute
#' @description Create a list of igraph objects from an edgelist according to a type attribute
#'
#' @param d data frame.
#' @param from column name of sender. If NULL, defaults to first column.
#' @param to column of receiver. If NULL, defaults to second column.
#' @param type type attribute to split the edgelist. If NULL, defaults to third column.
#' @param weight optional column name of edge weights. Ignored if NULL.
#' @param directed logical scalar, whether or not to create a directed graph.
#' @return list of igraph objects.
#' @author David Schoch
#' @examples
#' library(igraph)
#' d <- data.frame(
#'     from = rep(c(1, 2, 3), 3), to = rep(c(2, 3, 1), 3),
#'     type = rep(c("a", "b", "c"), each = 3), weight = 1:9
#' )
#' graph_from_multi_edgelist(d, "from", "to", "type", "weight")
#' @export

graph_from_multi_edgelist <- function(d, from = NULL, to = NULL, type = NULL, weight = NULL, directed = FALSE) {
    d <- as.data.frame(d)
    if (ncol(d) < 2) {
        stop("the data frame should contain at least three columns")
    }
    dnames <- names(d)

    if (!is.null(from) && !from %in% dnames) {
        stop(paste0(from, " is not a valid column name"))
    }
    if (!is.null(to) && !to %in% dnames) {
        stop(paste0(to, " is not a valid column name"))
    }
    if (!is.null(type) && !type %in% dnames) {
        stop(paste0(type, " is not a valid column name"))
    }

    if (is.null(from)) from <- dnames[1]
    if (is.null(to)) to <- dnames[2]
    if (is.null(type)) type <- dnames[3]

    if (is.null(weight)) {
        d <- d[, c(from, to, type)]
    } else {
        d <- d[, c(from, to, weight, type)]
    }

    d_lst <- split(d, d[[type]])
    g_lst <- lapply(d_lst, function(x) igraph::graph_from_data_frame(x, directed = directed))
    g_lst
}

#' @title k partite graphs
#' @description  Create a random k-partite graph.
#'
#' @param n number of nodes
#' @param grp vector of partition sizes
#' @return igraph object
#' @author David Schoch
#' @examples
#' # 3-partite graph with equal sized groups
#' graph_kpartite(n = 15, grp = c(5, 5, 5))
#' @export

graph_kpartite <- function(n = 10, grp = c(5, 5)) {
    g <- igraph::graph.empty(n = n, directed = FALSE)
    cur_node <- 1
    nodes <- 1:n
    for (i in 1:(length(grp) - 1)) {
        add_nodes <- cur_node:(cur_node + grp[i] - 1)
        add_edges <- c(t(expand.grid(add_nodes, nodes[nodes > max(add_nodes)])))
        g <- igraph::add_edges(g, add_edges)
        cur_node <- cur_node + grp[i]
    }
    return(g)
}

#' @title split graph
#' @description  Create a random split graph with a perfect core-periphery structure.
#'
#' @param n number of nodes
#' @param p probability of peripheral nodes to connect to the core nodes
#' @param core fraction of nodes in the core
#' @return igraph object
#' @author David Schoch
#' @examples
#' # split graph with 20 nodes and a core size of 10
#' split_graph(n = 20, p = 0.4, 0.5)
#' @export
split_graph <- function(n, p, core) {
    ncore <- floor(n * core)
    nperi <- n - ncore
    Acore <- matrix(1, ncore, ncore)
    Aperi <- (matrix(stats::runif(ncore * nperi), ncore, nperi) < p) + 0
    A <- rbind(cbind(Acore, Aperi), cbind(t(Aperi), matrix(0, nperi, nperi)))
    g <- igraph::graph_from_adjacency_matrix(A, "undirected", diag = FALSE)
    igraph::V(g)$core <- FALSE
    igraph::V(g)$core[1:ncore] <- TRUE
    g
}

test_that("bipartite_from_data_frame works", {
    library(igraph)
    df <- data.frame(from = c("a", "b", "c"), to = c("A", "A", "B"))
    g <- bipartite_from_data_frame(df, "from", "to")
    expect_equal(vcount(g) + ecount(g), 8)
})

test_that("bipartite_from_data_frame type1 error", {
    library(igraph)
    df <- data.frame(from = c("a", "b", "c"), to = c("A", "A", "B"))
    expect_error(bipartite_from_data_frame(df, "froom", "to"))
})

test_that("bipartite_from_data_frame overlap error", {
    library(igraph)
    df <- data.frame(from = c("a", "b", "c"), to = c("a", "A", "B"))
    expect_error(bipartite_from_data_frame(df, "from", "to"))
})

test_that("graph_from_multi_edgelist works", {
    library(igraph)
    d <- data.frame(
        from = rep(c(1, 2, 3), 3), to = rep(c(2, 3, 1), 3),
        type = rep(c("a", "b", "c"), each = 3), weight = 1:9
    )
    g <- graph_from_multi_edgelist(d, "from", "to", "type", "weight")
    expect_equal(length(g), 3)
})

test_that("graph_kpartite works", {
    library(igraph)
    g <- graph_kpartite(n = 15, grp = c(5, 5, 5))
    expect_equal(vcount(g) + ecount(g), 15 + 75)
})

#' @title two-mode network from a data.frame
#' @description Create a two-mode network from a data.frame
#'
#' @param d data.frame
#' @param type1 column name of mode 1
#' @param type2 column name of mode 2
#' @param attr named list of edge attributes
#' @param weighted should a weighted graph be created if multiple edges occur
#' @return two mode network as igraph object
#' @author David Schoch
#' @examples
#' library(igraph)
#' edges <- data.frame(mode1 = 1:5, mode2 = letters[1:5])
#' bipartite_from_data_frame(edges, "mode1", "mode2")
#' @export

bipartite_from_data_frame <- function(d, type1, type2, attr = NULL, weighted = TRUE) {
    if (!type1 %in% names(d)) {
        stop(paste0("no column named ", type1, " found in data frame"))
    }
    if (!type2 %in% names(d)) {
        stop(paste0("no column named ", type2, " found in data frame"))
    }

    mode1 <- unique(d[[type1]])
    mode2 <- unique(d[[type2]])
    if (any(mode1 %in% mode2)) {
        stop("some nodes appear in both modes. Modes in a two-mode network must be distinct.")
    }
    el <- cbind(d[[type1]], d[[type2]])

    g <- igraph::make_empty_graph(directed = FALSE)
    g <- igraph::add_vertices(g, nv = length(mode1), attr = list(name = mode1, type = TRUE))
    g <- igraph::add_vertices(g, nv = length(mode2), attr = list(name = mode2, type = FALSE))
    if (!is.null(attr)) {
        g <- igraph::add_edges(g, c(t(el)), attr = attr)
    } else {
        g <- igraph::add_edges(g, c(t(el)))
    }
    if (igraph::any_multiple(g) && weighted) {
        igraph::E(g)$weight <- 1
        g <- igraph::simplify(g,
            remove.multiple = TRUE,
            remove.loops = TRUE,
            edge.attr.comb = "sum"
        )
    }
    g
}

#' @title Multiple networks from a single edgelist with a typed attribute
#' @description Create a list of igraph objects from an edgelist according to a type attribute
#'
#' @param d data frame.
#' @param from column name of sender. If NULL, defaults to first column.
#' @param to column of receiver. If NULL, defaults to second column.
#' @param type type attribute to split the edgelist. If NULL, defaults to third column.
#' @param weight optional column name of edge weights. Ignored if NULL.
#' @param directed logical scalar, whether or not to create a directed graph.
#' @return list of igraph objects.
#' @author David Schoch
#' @examples
#' library(igraph)
#' d <- data.frame(
#'     from = rep(c(1, 2, 3), 3), to = rep(c(2, 3, 1), 3),
#'     type = rep(c("a", "b", "c"), each = 3), weight = 1:9
#' )
#' graph_from_multi_edgelist(d, "from", "to", "type", "weight")
#' @export

graph_from_multi_edgelist <- function(d, from = NULL, to = NULL, type = NULL, weight = NULL, directed = FALSE) {
    d <- as.data.frame(d)
    if (ncol(d) < 2) {
        stop("the data frame should contain at least three columns")
    }
    dnames <- names(d)

    if (!is.null(from) && !from %in% dnames) {
        stop(paste0(from, " is not a valid column name"))
    }
    if (!is.null(to) && !to %in% dnames) {
        stop(paste0(to, " is not a valid column name"))
    }
    if (!is.null(type) && !type %in% dnames) {
        stop(paste0(type, " is not a valid column name"))
    }

    if (is.null(from)) from <- dnames[1]
    if (is.null(to)) to <- dnames[2]
    if (is.null(type)) type <- dnames[3]

    if (is.null(weight)) {
        d <- d[, c(from, to, type)]
    } else {
        d <- d[, c(from, to, weight, type)]
    }

    d_lst <- split(d, d[[type]])
    g_lst <- lapply(d_lst, function(x) igraph::graph_from_data_frame(x, directed = directed))
    g_lst
}

#' @title k partite graphs
#' @description  Create a random k-partite graph.
#'
#' @param n number of nodes
#' @param grp vector of partition sizes
#' @return igraph object
#' @author David Schoch
#' @examples
#' # 3-partite graph with equal sized groups
#' graph_kpartite(n = 15, grp = c(5, 5, 5))
#' @export

graph_kpartite <- function(n = 10, grp = c(5, 5)) {
    g <- igraph::make_empty_graph(n = n, directed = FALSE)
    cur_node <- 1
    nodes <- seq_len(n)
    for (i in seq_len(length(grp) - 1)) {
        add_nodes <- cur_node:(cur_node + grp[i] - 1)
        add_edges <- c(t(expand.grid(add_nodes, nodes[nodes > max(add_nodes)])))
        g <- igraph::add_edges(g, add_edges)
        cur_node <- cur_node + grp[i]
    }
    return(g)
}

#' @title split graph
#' @description  Create a random split graph with a perfect core-periphery structure.
#'
#' @param n number of nodes
#' @param p probability of peripheral nodes to connect to the core nodes
#' @param core fraction of nodes in the core
#' @return igraph object
#' @author David Schoch
#' @examples
#' # split graph with 20 nodes and a core size of 10
#' split_graph(n = 20, p = 0.4, 0.5)
#' @export
split_graph <- function(n, p, core) {
    ncore <- floor(n * core)
    nperi <- n - ncore
    Acore <- matrix(1, ncore, ncore)
    Aperi <- (matrix(stats::runif(ncore * nperi), ncore, nperi) < p) + 0
    A <- rbind(cbind(Acore, Aperi), cbind(t(Aperi), matrix(0, nperi, nperi)))
    g <- igraph::graph_from_adjacency_matrix(A, "undirected", diag = FALSE)
    igraph::V(g)$core <- FALSE
    igraph::V(g)$core[1:ncore] <- TRUE
    g
}

#' @title Find Cliques, maximal or not, fast
#' @description Enumerates all (maximal) cliques using MACE. Can be faster than igraph in some circumstances
#' @param g An igraph object
#' @param what either "M" for maximal cliques or "C" for all cliques
#' @param min Numeric constant, lower limit on the size of the cliques to find. NULL means no limit, ie. it is the same as 0
#' @param max Numeric constant, upper limit on the size of the cliques to find. NULL means no limit
#' @param outfile character. If not NA, cliques are written to file
#' @details C Code downloaded from http://research.nii.ac.jp/~uno/codes.htm. Download the code and run make and then point an environment variable called MACE_PATH to the binary. See http://research.nii.ac.jp/~uno/code/mace.html for more details. MACE is faster than igraph for dense graphs.
#' @return a list containing numeric vectors of vertex ids. Each list element is a clique. If outfile!=NA, the output is written to the specified file
#' @author David Schoch
#' @references Kazuhisa Makino, Takeaki Uno, "New Algorithms for Enumerating All Maximal Cliques", Lecture Notes in Computer Science 3111 (Proceedings of SWAT 2004), Springer, pp.260-272, 2004
fast_cliques <- function(g, what = "M", min = NULL, max = NULL, outfile = NA) {
    if (!igraph::is_igraph(g)) {
        stop("g must be an igraph object")
    }
    what <- match.arg(what, c("M", "C"))
    what <- paste0(what, "_")
    ##########
    binary <- Sys.getenv("MACE_PATH")
    if (binary == "") {
        stop("MACE not found. Please set the MACE_PATH as environment variable.")
    }
    ##########
    fin <- tempfile()
    if (is.na(outfile)) {
        fout <- tempfile()
    } else {
        fout <- outfile
    }
    adj <- as_adj_list1(g)
    adj <- lapply(seq_along(adj), function(x) {
        neigh <- adj[[x]]
        neigh[neigh > x] - 1
    })
    writeLines(sapply(adj, paste, collapse = ","), fin)
    if (!is.null(min)) {
        what <- paste(what, "-l", min)
    }
    if (!is.null(max)) {
        what <- paste(what, "-u", max)
    }
    cmd <- paste(binary, what, fin, fout)
    a <- system(cmd, intern = TRUE)

    unlink(fin)
    if (is.na(outfile)) {
        clix <- readLines(fout)
        lst <- sapply(clix, function(x) strsplit(x, split = " "))
        lst <- lapply(lst, function(x) as.integer(x) + 1)
        lst <- unname(lst)
        unlink(fout)
        return(lst)
    } else {
        invisible(g)
    }
}

#' @title Find Cliques, maximal or not, fast
#' @description Enumerates all (maximal) cliques using MACE. Can be faster than igraph in some circumstances
#' @param g An igraph object
#' @param what either "M" for maximal cliques or "C" for all cliques
#' @param min Numeric constant, lower limit on the size of the cliques to find. NULL means no limit, ie. it is the same as 0
#' @param max Numeric constant, upper limit on the size of the cliques to find. NULL means no limit
#' @param outfile character. If not NA, cliques are written to file
#' @details C Code downloaded from http://research.nii.ac.jp/~uno/codes.htm. Download the code and run make and then point an environment variable called MACE_PATH to the binary. See http://research.nii.ac.jp/~uno/code/mace.html for more details. MACE is faster than igraph for dense graphs.
#' @return a list containing numeric vectors of vertex ids. Each list element is a clique. If outfile!=NA, the output is written to the specified file
#' @author David Schoch
#' @references Kazuhisa Makino, Takeaki Uno, "New Algorithms for Enumerating All Maximal Cliques", Lecture Notes in Computer Science 3111 (Proceedings of SWAT 2004), Springer, pp.260-272, 2004
fast_cliques <- function(g, what = "M", min = NULL, max = NULL, outfile = NA) {
    if (!igraph::is.igraph(g)) {
        stop("g must be an igraph object")
    }
    what <- match.arg(what, c("M", "C"))
    what <- paste0(what, "_")
    ##########
    binary <- Sys.getenv("MACE_PATH")
    if (binary == "") {
        stop("MACE not found. Please set the MACE_PATH as environment variable.")
    }
    ##########
    fin <- tempfile()
    if (is.na(outfile)) {
        fout <- tempfile()
    } else {
        fout <- outfile
    }
    adj <- as_adj_list1(g)
    adj <- lapply(1:length(adj), function(x) {
        neigh <- adj[[x]]
        neigh[neigh > x] - 1
    })
    writeLines(sapply(adj, paste, collapse = ","), fin)
    if (!is.null(min)) {
        what <- paste(what, "-l", min)
    }
    if (!is.null(max)) {
        what <- paste(what, "-u", max)
    }
    cmd <- paste(binary, what, fin, fout)
    a <- system(cmd, intern = TRUE)

    unlink(fin)
    if (is.na(outfile)) {
        clix <- readLines(fout)
        lst <- sapply(clix, function(x) strsplit(x, split = " "))
        lst <- lapply(lst, function(x) as.integer(x) + 1)
        lst <- unname(lst)
        unlink(fout)
        return(lst)
    } else {
        invisible(g)
    }
}

#' @title Cartesian product of two graphs
#' @description Compute the Cartesian product of two graphs
#' @param g An igraph object
#' @param h An igraph object
#' @details See https://en.wikipedia.org/wiki/Cartesian_product_of_graphs
#' @return Cartesian product as igraph object
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- make_ring(4)
#' h <- make_full_graph(2)
#' graph_cartesian(g, h)
#' @export
graph_cartesian <- function(g, h) {
    elg <- igraph::as_edgelist(g)
    elh <- igraph::as_edgelist(h)
    vg <- 1:igraph::vcount(g)
    vh <- 1:igraph::vcount(h)
    el_list <- vector("list", nrow(elg) + nrow(elh))
    for (i in seq_len(nrow(elg))) {
        el_list[[i]] <- matrix(apply(expand.grid(elg[i, ], vh), 1, function(x) paste(x, collapse = "-")), ncol = 2, byrow = TRUE)
    }
    for (i in seq_len(nrow(elh))) {
        el_list[[nrow(elg) + i]] <- matrix(apply(expand.grid(elh[i, ], vg), 1, function(x) paste(rev(x), collapse = "-")), ncol = 2, byrow = TRUE)
    }
    el <- do.call(rbind, el_list)
    igraph::graph_from_edgelist(el, F)
}

# direct graph product ----
#' @title Direct product of two graphs
#' @description Compute the direct product of two graphs
#' @param g An igraph object
#' @param h An igraph object
#' @details See https://en.wikipedia.org/wiki/Tensor_product_of_graphs
#' @return Direct product as igraph object
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- make_ring(4)
#' h <- make_full_graph(2)
#' graph_direct(g, h)
#' @export
graph_direct <- function(g, h) {
    elg <- igraph::as_edgelist(g)
    elh <- igraph::as_edgelist(h)
    el_list <- vector("list", nrow(elg) * nrow(elh) * 2)
    idx <- 1L
    for (i in seq_len(nrow(elg))) {
        for (j in seq_len(nrow(elh))) {
            el_list[[idx]] <- c(paste0(elg[i, 1], "-", elh[j, 1]), paste0(elg[i, 2], "-", elh[j, 2]))
            idx <- idx + 1L
            el_list[[idx]] <- c(paste0(elg[i, 2], "-", elh[j, 1]), paste0(elg[i, 1], "-", elh[j, 2]))
            idx <- idx + 1L
        }
    }
    el <- do.call(rbind, el_list)
    igraph::graph_from_edgelist(el, directed = FALSE)
}

test_that("graph products works", {
    library(igraph)
    g <- make_ring(4)
    h <- make_full_graph(2)
    gh <- graph_cartesian(g, h)
    expect_equal(20, vcount(gh) + ecount(gh))
    hg <- graph_direct(g, h)
    expect_equal(16, vcount(hg) + ecount(hg))
})

#' dyad census with node attributes
#'
#' @param g igraph object. should be a directed graph.
#' @param vattr name of vertex attribute to be used.
#' @return dyad census as a data.frame.
#' @details The node attribute should be integers from 1 to max(attr)
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- sample_gnp(10, 0.4, directed = TRUE)
#' V(g)$attr <- c(rep(1, 5), rep(2, 5))
#' dyad_census_attr(g, "attr")
#' @export
dyad_census_attr <- function(g, vattr) {
    if (!igraph::is_directed(g)) {
        stop("g must be a directed graph")
    }
    if (!vattr %in% igraph::vertex_attr_names(g)) {
        stop(paste0("there is no vertex attribute called ", vattr))
    }
    attr <- igraph::vertex_attr(g, vattr)
    if (!all(is.numeric(attr))) {
        stop("vertex attribute must be numeric ")
    }
    ns <- table(attr)

    df <- igraph::as_data_frame(g, "both")
    nodes <- df$vertices
    edges <- df$edges

    edges["from_attr"] <- nodes[[vattr]][edges[["from"]]]
    edges["to_attr"] <- nodes[[vattr]][edges[["to"]]]
    edges[["reci"]] <- igraph::which_mutual(g)
    edges[["count"]] <- 1

    empty_grid <- cbind(expand.grid(from_attr = 1:max(attr), to_attr = 1:max(attr)), count = 0)
    asym <- tryCatch(stats::aggregate(count ~ from_attr + to_attr, data = edges[!edges[["reci"]], ], FUN = sum),
        error = function(e) empty_grid
    )
    sym <- tryCatch(stats::aggregate(count ~ from_attr + to_attr, data = edges[edges[["reci"]], ], FUN = sum),
        error = function(e) empty_grid
    )

    idxx <- sym[["from_attr"]] == sym[["to_attr"]]
    sym[["count"]][idxx] <- sym[["count"]][idxx] / 2
    sym <- sym[sym[["from_attr"]] <= sym[["to_attr"]], ]

    idyx <- asym[["from_attr"]] > asym[["to_attr"]]
    asym2 <- asym[idyx, ]
    names(asym2) <- c(names(asym2)[2:1], "count2")
    asym <- merge(asym, asym2)
    names(asym)[3:4] <- c("asym_ab", "asym_ba")

    dc_df <- expand.grid(from_attr = 1:max(attr), to_attr = 1:max(attr))
    dc_df <- dc_df[dc_df[["from_attr"]] <= dc_df[["to_attr"]], ]
    dc_df <- merge(dc_df, asym, all.x = TRUE)
    dc_df[["asym_ab"]][is.na(dc_df[["asym_ab"]])] <- 0
    dc_df[["asym_ba"]][is.na(dc_df[["asym_ba"]])] <- 0

    dc_df <- merge(dc_df, sym, all.x = TRUE)
    names(dc_df)[5] <- "sym"
    dc_df[["sym"]][is.na(dc_df[["sym"]])] <- 0
    dc_df[["null"]] <- ifelse(dc_df[["from_attr"]] == dc_df[["to_attr"]],
        choose(ns[dc_df[["from_attr"]]], 2) - dc_df[["asym_ab"]] - dc_df[["asym_ba"]] - dc_df[["sym"]],
        ns[dc_df[["from_attr"]]] * ns[dc_df[["to_attr"]]] - dc_df[["asym_ab"]] - dc_df[["asym_ba"]] - dc_df[["sym"]]
    )
    dc_df
}

test_that("dyad census attr works", {
    library(igraph)
    gundir <- make_full_graph(3, directed = FALSE)
    gnoatt <- make_full_graph(3, directed = TRUE)
    gchatt <- make_full_graph(3, directed = TRUE)
    vertex_attr(gchatt, "test") <- c("a", "b", "c")
    gattr <- make_full_graph(3, directed = TRUE)
    vertex_attr(gattr, "test") <- c(1, 2, 3)
    expect_error(dyad_census_attr(gundir))
    expect_error(dyad_census_attr(gnoatt))
    expect_error(dyad_census_attr(gchatt, "test"))
    expect_equal(dyad_census_attr(gattr, "test")$sym, c(0, 1, 1, 0, 1, 0))
})

#' dyad census with node attributes
#'
#' @param g igraph object. should be a directed graph.
#' @param vattr name of vertex attribute to be used.
#' @return dyad census as a data.frame.
#' @details The node attribute should be integers from 1 to max(attr)
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- sample_gnp(10, 0.4, directed = TRUE)
#' V(g)$attr <- c(rep(1, 5), rep(2, 5))
#' dyad_census_attr(g, "attr")
#' @export
dyad_census_attr <- function(g, vattr) {
    attr <- validate_vattr(g, vattr)
    ns <- table(attr)

    df <- igraph::as_data_frame(g, "both")
    nodes <- df$vertices
    edges <- df$edges

    edges["from_attr"] <- nodes[[vattr]][edges[["from"]]]
    edges["to_attr"] <- nodes[[vattr]][edges[["to"]]]
    edges[["reci"]] <- igraph::which_mutual(g)
    edges[["count"]] <- 1

    empty_grid <- cbind(expand.grid(from_attr = 1:max(attr), to_attr = 1:max(attr)), count = 0)
    asym <- tryCatch(stats::aggregate(count ~ from_attr + to_attr, data = edges[!edges[["reci"]], ], FUN = sum),
        error = function(e) empty_grid
    )
    sym <- tryCatch(stats::aggregate(count ~ from_attr + to_attr, data = edges[edges[["reci"]], ], FUN = sum),
        error = function(e) empty_grid
    )

    idxx <- sym[["from_attr"]] == sym[["to_attr"]]
    sym[["count"]][idxx] <- sym[["count"]][idxx] / 2
    sym <- sym[sym[["from_attr"]] <= sym[["to_attr"]], ]

    idyx <- asym[["from_attr"]] > asym[["to_attr"]]
    asym2 <- asym[idyx, ]
    names(asym2) <- c(names(asym2)[2:1], "count2")
    asym <- merge(asym, asym2)
    names(asym)[3:4] <- c("asym_ab", "asym_ba")

    dc_df <- expand.grid(from_attr = 1:max(attr), to_attr = 1:max(attr))
    dc_df <- dc_df[dc_df[["from_attr"]] <= dc_df[["to_attr"]], ]
    dc_df <- merge(dc_df, asym, all.x = TRUE)
    dc_df[["asym_ab"]][is.na(dc_df[["asym_ab"]])] <- 0
    dc_df[["asym_ba"]][is.na(dc_df[["asym_ba"]])] <- 0

    dc_df <- merge(dc_df, sym, all.x = TRUE)
    names(dc_df)[5] <- "sym"
    dc_df[["sym"]][is.na(dc_df[["sym"]])] <- 0
    dc_df[["null"]] <- ifelse(dc_df[["from_attr"]] == dc_df[["to_attr"]],
        choose(ns[dc_df[["from_attr"]]], 2) - dc_df[["asym_ab"]] - dc_df[["asym_ba"]] - dc_df[["sym"]],
        ns[dc_df[["from_attr"]]] * ns[dc_df[["to_attr"]]] - dc_df[["asym_ab"]] - dc_df[["asym_ba"]] - dc_df[["sym"]]
    )
    dc_df
}

#' @title helper function
#' @description small functions to deal with typical network problems
#'
#' @param g igraph object
#' @name helpers
#' @return igraph object
#' @author David Schoch
NULL

#' @rdname helpers
#' @export
biggest_component <- function(g) {
    comps <- igraph::components(g, mode = "weak")
    igraph::induced_subgraph(g, which(comps$membership == which.max(comps$csize)))
}

#' @rdname helpers
#' @export
delete_isolates <- function(g) {
    igraph::delete_vertices(g, which(igraph::degree(g) == 0))
}

test_that("biggest_component works", {
  library(igraph)
  g <- make_full_graph(4) + make_full_graph(2)
  bg <- biggest_component(g)
  expect_equal(vcount(bg),4L)
})

test_that("delete_isolates works", {
  library(igraph)
  g <- make_full_graph(4) + make_full_graph(1)
  bg <- delete_isolates(g)
  expect_equal(vcount(bg),4L)
})

#' @title helper function
#' @description small functions to deal with typical network problems
#'
#' @param g igraph object
#' @name helpers
#' @return igraph object
#' @author David Schoch
NULL

#' @rdname helpers
#' @export
biggest_component <- function(g) {
    comps <- igraph::components(g, mode = "weak")
    igraph::induced_subgraph(g, which(comps$membership == which.max(comps$csize)))
}

#' @rdname helpers
#' @export
delete_isolates <- function(g) {
    igraph::delete_vertices(g, which(igraph::degree(g) == 0))
}

# validate that `g` is a directed graph carrying a numeric vertex attribute
# `vattr`, and return that attribute vector. Used by the *_census_attr functions.
validate_vattr <- function(g, vattr) {
    if (!igraph::is_directed(g)) {
        stop("g must be a directed graph", call. = FALSE)
    }
    if (!vattr %in% igraph::vertex_attr_names(g)) {
        stop("there is no vertex attribute called ", vattr, call. = FALSE)
    }
    attr <- igraph::vertex_attr(g, vattr)
    if (!all(is.numeric(attr))) {
        stop("vertex attribute must be numeric ", call. = FALSE)
    }
    attr
}

#' Reciprocity correlation coefficient
#'
#' @param g igraph object. should be a directed graph
#' @return Reciprocity as a correlation
#' @details
#' The usual definition of reciprocity has some defects. It cannot tell the relative difference of reciprocity compared with purely random network with the same number of vertices and edges.
#' The useful information from reciprocity is not the value itself, but whether mutual links occur more or less often than expected by chance.
#'
#' To overcome this issue, reciprocity can be defined as the correlation coefficient between the entries of the adjacency matrix of a directed graph:
#' \deqn{
#'  \frac{\sum_{i\neq j} (a_{ij} - a')((a_{ji} - a')}{\sum_{i\neq j} (a_{ij} - a')^2}
#' }
#' where a' is the density of g.
#'
#' This definition gives an absolute quantity which directly allows one to distinguish between reciprocal (>0) and antireciprocal (< 0) networks, with mutual links occurring more and less often than random respectively.
#' @references Diego Garlaschelli; Loffredo, Maria I. (2004). "Patterns of Link Reciprocity in Directed Networks". Physical Review Letters. American Physical Society. 93 (26): 268701
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- sample_gnp(20, p = 0.3, directed = TRUE)
#' reciprocity(g)
#' reciprocity_cor(g)
#' @export
reciprocity_cor <- function(g) {
    if (!igraph::is_igraph(g)) {
        stop("g must be an igraph object")
    }
    if (!igraph::is_directed(g)) {
        stop("g must be directed")
    }
    r <- igraph::reciprocity(g)
    d <- igraph::edge_density(g)
    if (d == 1) {
        return(0)
    } else {
        (r - d) / (1 - d)
    }
}

library(igraph)
test_that("reciprocity_cor works", {
    g <- make_full_graph(5, directed = TRUE)
    expect_equal(reciprocity_cor(g), 0)
})


test_that("reciprocity_cor non igraph error", {
    g <- 42
    expect_error(reciprocity_cor(g))
})

test_that("reciprocity_cor directed error", {
    g <- make_full_graph(5, directed = FALSE)
    expect_error(reciprocity_cor(g))
})

#' @title Adjacency list
#' @description Create adjacency lists from a graph, either for adjacent edges or for neighboring vertices. This version is faster than the version of igraph but less general.
#' @param g An igraph object
#' @details The function does not have a mode parameter and only returns the adjacency list comparable to as_adj_list(g,mode="all)
#' @return A list of numeric vectors.
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- make_ring(10)
#' as_adj_list1(g)
#' @export
as_adj_list1 <- function(g) {
    n <- igraph::vcount(g)
    lapply(1:n, function(i) {
        x <- g[[i]][[1]]
        attr(x, "env") <- NULL
        attr(x, "graph") <- NULL
        class(x) <- NULL
        x
    })
}

#' @title weighted dense adjacency matrix
#' @description returns the weighted adjacency matrix in dense format
#' @param g An igraph object
#' @param attr Either NULL or a character string giving an edge attribute name. If NULL a traditional adjacency matrix is returned. If not NULL then the values of the given edge attribute are included in the adjacency matrix.
#' @details This method is faster than as_adj from igraph if you need the weighted adjacency matrix in dense format
#' @return Numeric matrix
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- sample_gnp(10, 0.2)
#' E(g)$weight <- runif(ecount(g))
#' as_adj_weighted(g, attr = "weight")
#' @export
as_adj_weighted <- function(g, attr = NULL) {
    as.matrix(igraph::as_adjacency_matrix(
        g,
        attr = attr,
        type = "both",
        sparse = TRUE
    ))
}


#' @title Clique Vertex Matrix
#' @description Creates the clique vertex matrix with entries (i,j) equal to one if node j is in clique i
#' @param g An igraph object
#' @return Numeric matrix
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- sample_gnp(10, 0.2)
#' clique_vertex_mat(g)
#' @export
clique_vertex_mat <- function(g) {
    if (!igraph::is_igraph(g)) {
        stop("g must be an igraph object")
    }
    if (igraph::is_directed(g)) {
        warning("g is directed. Underlying undirected graph is used")
        g <- igraph::as.undirected(g)
    }
    mcl <- igraph::max_cliques(g)
    M <- matrix(0, length(mcl), igraph::vcount(g))
    for (i in seq_len(length(mcl))) {
        M[i, mcl[[i]]] <- 1
    }
    M
}


#' @title Convert a list of graphs to an adjacency matrices
#' @description Convenience function that turns a list of igraph objects into adjacency matrices.
#' @param g_lst A list of igraph object
#' @param attr Either NULL or a character string giving an edge attribute name. If NULL a binary adjacency matrix is returned.
#' @param sparse Logical scalar, whether to create a sparse matrix. The 'Matrix' package must be installed for creating sparse matrices.
#' @return List of numeric matrices
#' @author David Schoch
#' @export
as_multi_adj <- function(g_lst, attr = NULL, sparse = FALSE) {
    if (!all(unlist(lapply(g_lst, igraph::is.igraph)))) {
        stop("all entries of g_lst must be igraph objects")
    }
    lapply(g_lst, function(x) {
        igraph::as_adjacency_matrix(x, "both", attr = attr, sparse = sparse)
    })
}

test_that("as_adj_list1 works", {
    library(igraph)
    l <- list(c(2, 3), c(1), c(1))
    g <- graph_from_adj_list(l, mode = "all")
    expect_equal(l, as_adj_list1(g))
})

test_that("as_adj_weighted works", {
    library(igraph)
    A <- matrix(c(0, 3, 3, 3, 0, 3, 3, 3, 0), 3, 3)
    g <- graph_from_adjacency_matrix(A, mode = "undirected", weighted = TRUE)
    expect_equal(as_adj_weighted(g, attr = "weight")[1, 2], 3)
})

test_that("clique_vertex_mat works", {
    library(igraph)
    g <- make_full_graph(3)
    expect_equal(clique_vertex_mat(g), matrix(1, ncol = 3, nrow = 1))
})

#' Generate random graphs with a given coreness sequence
#' @description Similar to \link[igraph]{sample_degseq} just with \link[igraph]{coreness}
#' @param cores coreness sequence
#' @details The code is an adaption of the python code from https://github.com/ktvank/Random-Graphs-with-Prescribed-K-Core-Sequences/
#' @return igraph object of graph with the same coreness sequence as the input
#' @references
#' Van Koevering, Katherine, Austin R. Benson, and Jon Kleinberg. 2021. ‘Random Graphs with Prescribed K-Core Sequences: A New Null Model for Network Analysis’. ArXiv:2102.12604. https://doi.org/10.1145/3442381.3450001.
#' @author David Schoch
#' @examples
#' library(igraph)
#' g1 <- make_graph("Zachary")
#' kcores1 <- coreness(g1)
#' g2 <- sample_coreseq(kcores1)
#' kcores2 <- coreness(g2)
#'
#' # the sorted arrays are the same
#' all(sort(kcores1) == sort(kcores2))
#'
#' @export
sample_coreseq <- function(cores) {
    cores <- sort(cores, decreasing = TRUE)

    if (!is_kcoreseq(cores)) {
        stop("sequence is not a valid kcore sequence")
    }
    K <- cores[1]

    indices <- lapply(0:K, function(x) which(cores == x))

    if (K == 0) {
        n0 <- length(indices[[1]])
        g <- igraph::make_empty_graph(n = n0, directed = FALSE)
        return(g)
    } else if (K == 1) {
        n0 <- length(indices[[1]])
        n1 <- length(indices[[2]])
        g <- igraph::make_empty_graph(n = n1, directed = FALSE)
        g <- igraph::add_edges(g, edges = c(t(cbind(1:(n1 - 1), 2:n1))))
        g <- igraph::add_vertices(g, nv = n0)
        return(g)
    } else if (K == 2) {
        n0 <- length(indices[[1]])
        n1 <- length(indices[[2]])
        n2 <- length(indices[[3]])
        g <- igraph::graph.empty(n = n2, directed = FALSE)
        g <- igraph::add_edges(g, edges = c(t(cbind(1:(n2 - 1), 2:n2))))
        g <- igraph::add_edges(g, edges = c(1, n2))
        g <- add_lower_nodes(cores, indices, g, 1)
        return(g)
    } else {
        g <- igraph::graph.empty(directed = FALSE)
        g <- generate_k_graph(K, sum(cores == K), g)
        g <- add_lower_nodes(cores, indices, g, K - 1)
        return(g)
    }
}

is_kcoreseq <- function(cores) {
    cores <- sort(cores, decreasing = TRUE)
    num_max_core_val <- sum(cores == max(cores))
    return(num_max_core_val >= cores[1])
}

add_lower_nodes <- function(cores, indices, g, k) {
    K <- cores[1]
    higher_nodes <- sort(unname(unlist(indices[(k + 2):(K + 1)])))
    g <- igraph::add_vertices(g, nv = length(indices[[k + 1]]))
    if (k == 0) {
        return(g)
    }
    all_edges <- integer(0)
    for (v in indices[[k + 1]]) {
        randv <- sample(higher_nodes, k, replace = FALSE)
        all_edges <- c(all_edges, c(t(cbind(randv, v))))
    }
    g <- igraph::add_edges(g, all_edges)
    return(add_lower_nodes(cores, indices, g, k - 1))
}

generate_k_graph <- function(C, N, g) {
    if (!((C %% 2) == (N %% 2))) {
        g <- igraph::add_vertices(g, nv = N)
        g <- igraph::add_edges(g, edges = c(t(cbind(1:(N - 1), 2:N))))
        g <- igraph::add_edges(g, edges = c(1, N))
        z <- ceiling((N - C + 1) / 2)
        all_edges <- integer(0)
        for (i in 0:(N - 1)) {
            start <- (i + z + 1) %% (N) # BUG POTENTIAL!!!
            stop <- (i - z) %% (N) # BUG POTENTIAL!!!

            if (stop >= start) {
                listv <- start:(stop - 1) # BUG POTENTIAL!!!
                edges <- c(t(cbind(listv, i)))
            } else {
                listv <- c((0:(N - 1))[0:(stop)], (0:(N - 1))[(start + 1):N])
                edges <- c(t(cbind(listv, i)))
            }
            all_edges <- c(all_edges, edges + 1)
        }
        g <- igraph::add_edges(g, all_edges)
        g <- igraph::simplify(g)
    } else {
        g <- generate_k_graph(C, N - 1, g)
        g <- igraph::add_vertices(g, 1)
        randv <- sample(1:(N - 1), C)
        g <- igraph::add_edges(g, c(t(cbind(randv, N))))
    }
    return(g)
}

test_that("kcore sampling works", {
    library(igraph)
    cores00 <- c(0, 0, 0)
    cores01 <- c(1, 1, 1)
    cores02 <- c(2, 2, 2)
    coresNA <- c(4, 2, 2)
    coresCl <- c(4, 4, 4, 4, 4)
    expect_equal(coreness(sample_coreseq(cores00)), cores00)
    expect_equal(coreness(sample_coreseq(cores01)), cores01)
    expect_equal(coreness(sample_coreseq(cores02)), cores02)
    expect_error(sample_coreseq(coresNA))
    expect_equal(coreness(sample_coreseq(coresCl)), coresCl)
})

#' @title Adjacency list
#' @description Create adjacency lists from a graph, either for adjacent edges or for neighboring vertices. This version is faster than the version of igraph but less general.
#' @param g An igraph object
#' @details The function does not have a mode parameter and only returns the adjacency list comparable to as_adj_list(g,mode="all)
#' @return A list of numeric vectors.
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- make_ring(10)
#' as_adj_list1(g)
#' @export
as_adj_list1 <- function(g) {
    n <- igraph::vcount(g)
    lapply(seq_len(n), function(i) {
        x <- g[[i]][[1]]
        attr(x, "env") <- NULL
        attr(x, "graph") <- NULL
        class(x) <- NULL
        x
    })
}

#' @title weighted dense adjacency matrix
#' @description returns the weighted adjacency matrix in dense format
#' @param g An igraph object
#' @param attr Either NULL or a character string giving an edge attribute name. If NULL a traditional adjacency matrix is returned. If not NULL then the values of the given edge attribute are included in the adjacency matrix.
#' @details This method is faster than as_adj from igraph if you need the weighted adjacency matrix in dense format
#' @return Numeric matrix
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- sample_gnp(10, 0.2)
#' E(g)$weight <- runif(ecount(g))
#' as_adj_weighted(g, attr = "weight")
#' @export
as_adj_weighted <- function(g, attr = NULL) {
    as.matrix(igraph::as_adjacency_matrix(
        g,
        attr = attr,
        type = "both",
        sparse = TRUE
    ))
}


#' @title Clique Vertex Matrix
#' @description Creates the clique vertex matrix with entries (i,j) equal to one if node j is in clique i
#' @param g An igraph object
#' @return Numeric matrix
#' @author David Schoch
#' @examples
#' library(igraph)
#' g <- sample_gnp(10, 0.2)
#' clique_vertex_mat(g)
#' @export
clique_vertex_mat <- function(g) {
    if (!igraph::is_igraph(g)) {
        stop("g must be an igraph object")
    }
    if (igraph::is_directed(g)) {
        warning("g is directed. Underlying undirected graph is used")
        g <- igraph::as_undirected(g)
    }
    mcl <- igraph::max_cliques(g)
    M <- matrix(0, length(mcl), igraph::vcount(g))
    for (i in seq_len(length(mcl))) {
        M[i, mcl[[i]]] <- 1
    }
    M
}


#' @title Convert a list of graphs to an adjacency matrices
#' @description Convenience function that turns a list of igraph objects into adjacency matrices.
#' @param g_lst A list of igraph object
#' @param attr Either NULL or a character string giving an edge attribute name. If NULL a binary adjacency matrix is returned.
#' @param sparse Logical scalar, whether to create a sparse matrix. The 'Matrix' package must be installed for creating sparse matrices.
#' @return List of numeric matrices
#' @author David Schoch
#' @export
as_multi_adj <- function(g_lst, attr = NULL, sparse = FALSE) {
    if (!all(unlist(lapply(g_lst, igraph::is_igraph)))) {
        stop("all entries of g_lst must be igraph objects")
    }
    lapply(g_lst, function(x) {
        igraph::as_adjacency_matrix(x, "both", attr = attr, sparse = sparse)
    })
}

#' Generate random graphs with a given coreness sequence
#' @description Similar to \link[igraph]{sample_degseq} just with \link[igraph]{coreness}
#' @param cores coreness sequence
#' @details The code is an adaption of the python code from https://github.com/ktvank/Random-Graphs-with-Prescribed-K-Core-Sequences/
#' @return igraph object of graph with the same coreness sequence as the input
#' @references
#' Van Koevering, Katherine, Austin R. Benson, and Jon Kleinberg. 2021. ‘Random Graphs with Prescribed K-Core Sequences: A New Null Model for Network Analysis’. ArXiv:2102.12604. https://doi.org/10.1145/3442381.3450001.
#' @author David Schoch
#' @examples
#' library(igraph)
#' g1 <- make_graph("Zachary")
#' kcores1 <- coreness(g1)
#' g2 <- sample_coreseq(kcores1)
#' kcores2 <- coreness(g2)
#'
#' # the sorted arrays are the same
#' all(sort(kcores1) == sort(kcores2))
#'
#' @export
sample_coreseq <- function(cores) {
    cores <- sort(cores, decreasing = TRUE)

    if (!is_kcoreseq(cores)) {
        stop("sequence is not a valid kcore sequence")
    }
    K <- cores[1]

    indices <- lapply(0:K, function(x) which(cores == x))

    if (K == 0) {
        n0 <- length(indices[[1]])
        g <- igraph::make_empty_graph(n = n0, directed = FALSE)
        return(g)
    } else if (K == 1) {
        n0 <- length(indices[[1]])
        n1 <- length(indices[[2]])
        g <- igraph::make_empty_graph(n = n1, directed = FALSE)
        g <- igraph::add_edges(g, edges = c(t(cbind(1:(n1 - 1), 2:n1))))
        g <- igraph::add_vertices(g, nv = n0)
        return(g)
    } else if (K == 2) {
        n0 <- length(indices[[1]])
        n1 <- length(indices[[2]])
        n2 <- length(indices[[3]])
        g <- igraph::make_empty_graph(n = n2, directed = FALSE)
        g <- igraph::add_edges(g, edges = c(t(cbind(1:(n2 - 1), 2:n2))))
        g <- igraph::add_edges(g, edges = c(1, n2))
        g <- add_lower_nodes(cores, indices, g, 1)
        return(g)
    } else {
        g <- igraph::make_empty_graph(directed = FALSE)
        g <- generate_k_graph(K, sum(cores == K), g)
        g <- add_lower_nodes(cores, indices, g, K - 1)
        return(g)
    }
}

is_kcoreseq <- function(cores) {
    cores <- sort(cores, decreasing = TRUE)
    num_max_core_val <- sum(cores == max(cores))
    return(num_max_core_val >= cores[1])
}

add_lower_nodes <- function(cores, indices, g, k) {
    K <- cores[1]
    higher_nodes <- sort(unname(unlist(indices[(k + 2):(K + 1)])))
    g <- igraph::add_vertices(g, nv = length(indices[[k + 1]]))
    if (k == 0) {
        return(g)
    }
    all_edges <- integer(0)
    for (v in indices[[k + 1]]) {
        randv <- sample(higher_nodes, k, replace = FALSE)
        all_edges <- c(all_edges, c(t(cbind(randv, v))))
    }
    g <- igraph::add_edges(g, all_edges)
    return(add_lower_nodes(cores, indices, g, k - 1))
}

generate_k_graph <- function(C, N, g) {
    if (!((C %% 2) == (N %% 2))) {
        g <- igraph::add_vertices(g, nv = N)
        g <- igraph::add_edges(g, edges = c(t(cbind(1:(N - 1), 2:N))))
        g <- igraph::add_edges(g, edges = c(1, N))
        z <- ceiling((N - C + 1) / 2)
        all_edges <- integer(0)
        for (i in 0:(N - 1)) {
            start <- (i + z + 1) %% (N) # BUG POTENTIAL!!!
            stop <- (i - z) %% (N) # BUG POTENTIAL!!!

            if (stop >= start) {
                listv <- start:(stop - 1) # BUG POTENTIAL!!!
                edges <- c(t(cbind(listv, i)))
            } else {
                listv <- c((0:(N - 1))[0:(stop)], (0:(N - 1))[(start + 1):N])
                edges <- c(t(cbind(listv, i)))
            }
            all_edges <- c(all_edges, edges + 1)
        }
        g <- igraph::add_edges(g, all_edges)
        g <- igraph::simplify(g)
    } else {
        g <- generate_k_graph(C, N - 1, g)
        g <- igraph::add_vertices(g, 1)
        randv <- sample(1:(N - 1), C)
        g <- igraph::add_edges(g, c(t(cbind(randv, N))))
    }
    return(g)
}

#' Homophilic random graph using BA preferential attachment model
#' @description A graph of n nodes is grown by attaching new nodes each with m
#' edges that are preferentially attached to existing nodes with high
#' degree, depending on the homophily parameters.
#' @param n number of nodes
#' @param m number of edges a new node is connected to
#' @param minority_fraction fraction of nodes that belong to the minority group
#' @param h_ab probability to connect a node from group a with groub b
#' @param h_ba probability to connect a node from group b with groub a. If NULL, h_ab is used.
#' @param directed should a directed network be created
#' @details The code is an adaption of the python code from https://github.com/gesiscss/HomophilicNtwMinorities/
#' @return igraph object
#' @references
#' Karimi, F., Génois, M., Wagner, C., Singer, P., & Strohmaier, M. (2018). Homophily influences ranking of minorities in social networks. Scientific reports, 8(1), 1-12. (https://www.nature.com/articles/s41598-018-29405-7)
#'
#' Espín-Noboa, L., Wagner, C., Strohmaier, M., & Karimi, F. (2022). Inequality and inequity in network-based ranking and recommendation algorithms. Scientific reports, 12(1), 1-14. (https://www.nature.com/articles/s41598-022-05434-1)
#' @author David Schoch
#' #maximally heterophilic network
#' sample_pa_homophilic(n = 50, m = 2,minority_fraction = 0.2,h_ab = 1)
#' #maximally homophilic network
#' sample_pa_homophilic(n = 50, m = 2,minority_fraction = 0.2,h_ab = 0)
#' @export
sample_pa_homophilic <- function(n, m, minority_fraction, h_ab, h_ba = NULL, directed = FALSE) {
    if (is.null(h_ba)) {
        h_ba <- h_ab
    }
    h_aa <- 1 - h_ab
    h_bb <- 1 - h_ba
    minority_attr <- sample(
        c(
            rep(TRUE, floor(minority_fraction * n)),
            rep(FALSE, n - floor(minority_fraction * n))
        )
    )

    g <- igraph::make_empty_graph(n = 0, directed = directed)
    g <- igraph::add_vertices(g, n, attr = list(minority = minority_attr))

    dist <- matrix(NA, n, n)
    dist[outer(minority_attr, minority_attr, "&")] <- h_aa # within minority
    dist[outer(!minority_attr, !minority_attr, "&")] <- h_bb # within majority
    dist[outer(minority_attr, !minority_attr, "&")] <- h_ab # min->maj
    dist[outer(!minority_attr, minority_attr, "&")] <- h_ba # maj->min


    target_list <- seq_len(m)
    source <- m + 1
    while (source <= n) {
        deg <- igraph::degree(g)
        targets <- pick_targets(deg, source, target_list, dist, m)
        if (length(targets != 0)) {
            el <- rbind(source, targets)
            g <- igraph::add_edges(g, c(el))
        }
        target_list <- c(target_list, source)
        source <- source + 1
    }
    return(g)
}

pick_targets <- function(deg, source, target_list, dist, m) {
    target_prob <- dist[source, target_list] * (deg[target_list] + 1e-5)

    if (sum(target_prob > 0) < m) {
        return(c())
    } else {
        targets <- sample(target_list, m, prob = target_prob)
        return(targets)
    }
}

test_that("sample_pa_homophilic works", {
    g <- sample_pa_homophilic(
        n = 50,
        m = 2,
        minority_fraction = 0.2,
        h_ab = 1
    )
    expect_true(sum(V(g)$minority) == 10)
})

#' Graph correlation
#'
#' This function computes the correlation between networks. Implemented methods expect
#' the graph to be an adjacency matrix, an igraph, or a network object.
#'
#' @param object1  igraph object or adjacency matrix
#' @param object2  igraph object or adjacency matrix over the same vertex set as object1
#' @param ... additional arguments
#'
#' @return correlation between graphs
#'
#' @export
graph_cor <- function(object1, object2) UseMethod("graph_cor")

#' @rdname graph_cor
#' @method graph_cor default
#' @export
graph_cor.default <- function(object1, object2) {
    stop(
        "don't know how to handle class ",
        dQuote(data.class(object1)),
        " and ",
        dQuote(data.class(object2))
    )
}

#' @rdname graph_cor
#' @method graph_cor igraph
#' @export
graph_cor.igraph <- function(object1, object2, ...) {
    A1 <- igraph::as_adjacency_matrix(
        object1,
        type = "both",
        sparse = FALSE,
        ...
    )
    A2 <- igraph::as_adjacency_matrix(
        object2,
        type = "both",
        sparse = FALSE,
        ...
    )
    graph_cor.matrix(A1, A2)
}

#' @rdname graph_cor
#' @method graph_cor matrix
#' @export
graph_cor.matrix <- function(object1, object2) {
    stats::cor(c(object1), c(object2), use = "complete.obs")
}

#' @rdname graph_cor
#' @method graph_cor array
#' @export
graph_cor.array <- function(object1, object2) {
    stats::cor(c(object1), c(object2), use = "complete.obs")
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

benchmark <- function(excess, defect, num_nodes, average_k, max_degree, tau, tau2, mixing_parameter, overlapping_nodes, overlap_membership, nmin, nmax, fixed_range) {
    .Call(`_netUtils_benchmark`, excess, defect, num_nodes, average_k, max_degree, tau, tau2, mixing_parameter, overlapping_nodes, overlap_membership, nmin, nmax, fixed_range)
}

mse <- function(adjList, deg) {
    .Call(`_netUtils_mse`, adjList, deg)
}

sortxy <- function(x, y) {
    .Call(`_netUtils_sortxy`, x, y)
}

triadCensusCol <- function(A, attr, orbitClasses, triads) {
    .Call(`_netUtils_triadCensusCol`, A, attr, orbitClasses, triads)
}

#' Graph correlation
#'
#' This function computes the correlation between networks. Implemented methods expect
#' the graph to be an adjacency matrix, an igraph, or a network object.
#'
#' @param object1  igraph object or adjacency matrix
#' @param object2  igraph object or adjacency matrix over the same vertex set as object1
#' @param ... additional arguments
#'
#' @return correlation between graphs
#'
#' @export
graph_cor <- function(object1, object2) UseMethod("graph_cor")

#' @rdname graph_cor
#' @method graph_cor default
#' @export
graph_cor.default <- function(object1, object2) {
    stop(
        "don't know how to handle class ",
        dQuote(data.class(object1)),
        " and ",
        dQuote(data.class(object2))
    )
}

#' @rdname graph_cor
#' @method graph_cor igraph
#' @export
graph_cor.igraph <- function(object1, object2, ...) {
    A1 <- igraph::as_adjacency_matrix(
        object1,
        type = "both",
        sparse = FALSE,
        ...
    )
    A2 <- igraph::as_adjacency_matrix(
        object2,
        type = "both",
        sparse = FALSE,
        ...
    )
    graph_cor.matrix(A1, A2)
}

#' @rdname graph_cor
#' @method graph_cor matrix
#' @export
graph_cor.matrix <- function(object1, object2) {
    stats::cor(c(object1), c(object2), use = "complete.obs")
}

#' @rdname graph_cor
#' @method graph_cor array
#' @export
graph_cor.array <- graph_cor.matrix

#' @title Print graphs to terminal
#' @description Prints an igraph object to terminal (different than the standard igraph method)
#' @param object An igraph object
#' @param ... additional arguments to print (ignored)
#' @return str does not return anything. The obvious side effect is output to the terminal.
#' @author David Schoch
#' @export
str.igraph <- function(object, ...) {
    maxpr <- getOption("width")
    gattrs <- igraph::graph.attributes(object)
    vattrs <- igraph::vertex.attributes(object)
    eattrs <- igraph::edge.attributes(object)
    comps <- igraph::components(object)
    dens <- igraph::graph.density(object)

    # header
    if (!"name" %in% names(gattrs)) {
        gname <- "Unnamed Network"
    } else {
        gname <- gattrs[["name"]]
    }

    head <- paste(toupper(gname), " ",
        c("(undirected", "(directed")[igraph::is.directed(object) + 1], ", ",
        c("unweighted", "weighted")[igraph::is.weighted(object) + 1], ", ",
        c("", "signed, ")["sign" %in% names(eattrs) + 1],
        c("one-mode", "two-mode")[igraph::is.bipartite(object) + 1], " ",
        "network)",
        sep = ""
    )

    head <- paste0(head, "\n")
    delim <- paste0(rep("-", min(c(nchar(head), maxpr))), collapse = "")
    delim <- paste0(delim, "\n")
    short_delim <- "---\n"

    # graph details
    n <- paste0("Nodes: ", igraph::vcount(object))
    m <- paste0("Edges: ", igraph::ecount(object))
    d <- paste0("Density: ", ifelse(dens < 1e-4, "<1e-4", round(dens, 4)))
    cc <- paste0("Components: ", comps$no)
    iso <- paste0("Isolates: ", sum(igraph::degree(object) == 0))
    gstats <- paste(n, m, d, cc, iso, sep = ", ")
    gstats <- paste0(strwrap(gstats), collapse = "\n")
    gstats <- paste0(gstats, "\n")

    # graph attrs
    gattr_str <- ""
    if (length(gattrs) > 0) {
        gattr_str <- apply(cbind(names(gattrs), paste0("(", substr(sapply(gattrs, mode), 1, 1), ")")), 1, paste0, collapse = "")
        gattr_str <- paste("-Graph Attributes:\n ", paste0(gattr_str, collapse = ", "))
        gattr_str <- paste0(gattr_str, "\n", short_delim)
    }

    vattr_str <- format_attr_section(vattrs, "Vertex", short_delim)
    eattr_str <- format_attr_section(eattrs, "Edge", short_delim)
    if (igraph::ecount(object) > 0) {
        edges <- igraph::get.edgelist(object)[1:min(c(10, igraph::ecount(object))), ]
        edges <- strwrap(paste0(apply(edges, 1, paste0, collapse = c("--", "->")[igraph::is.directed(object) + 1]), collapse = " "))
        edges <- paste(edges, collapse = "\n")
        if (igraph::ecount(object) > 10) {
            edges <- c("-Edges (first 10): \n ", edges)
        } else {
            edges <- c("-Edges: \n ", edges)
        }
    } else {
        edges <- ""
    }
    cat(delim, head, delim, gstats, gattr_str, vattr_str, eattr_str, edges, sep = "")
}


format_attr_section <- function(attrs, label, short_delim) {
    if (length(attrs) == 0) return("")
    lines <- vapply(seq_along(attrs), function(l) {
        peek <- head_dot(attrs[[l]], nchar(names(attrs)[l]) + 4)
        prefix <- if (l == 1) "" else " "
        paste0(prefix, names(attrs)[l], "(", substr(mode(attrs[[l]]), 1, 1), "): ", peek, "\n")
    }, character(1))
    paste0("-", label, " Attributes:\n ", paste0(lines, collapse = ""), short_delim)
}

head_dot <- function(x, lname) {
    stri <- strwrap(paste0(x, collapse = ", "), width = 0.9 * getOption("width") - lname)[1]
    paste0(stri, " ...")
}

#' @title Print graphs to terminal
#' @description Prints an igraph object to terminal (different than the standard igraph method)
#' @param object An igraph object
#' @param ... additional arguments to print (ignored)
#' @return str does not return anything. The obvious side effect is output to the terminal.
#' @author David Schoch
#' @export
str.igraph <- function(object, ...) {
    maxpr <- getOption("width")
    gattrs <- igraph::graph_attr(object)
    vattrs <- igraph::vertex_attr(object)
    eattrs <- igraph::edge_attr(object)
    comps <- igraph::components(object)
    dens <- igraph::edge_density(object)

    # header
    if (!"name" %in% names(gattrs)) {
        gname <- "Unnamed Network"
    } else {
        gname <- gattrs[["name"]]
    }

    head <- paste(toupper(gname), " ",
        c("(undirected", "(directed")[igraph::is_directed(object) + 1], ", ",
        c("unweighted", "weighted")[igraph::is_weighted(object) + 1], ", ",
        c("", "signed, ")["sign" %in% names(eattrs) + 1],
        c("one-mode", "two-mode")[igraph::is_bipartite(object) + 1], " ",
        "network)",
        sep = ""
    )

    head <- paste0(head, "\n")
    delim <- paste0(rep("-", min(c(nchar(head), maxpr))), collapse = "")
    delim <- paste0(delim, "\n")
    short_delim <- "---\n"

    # graph details
    n <- paste0("Nodes: ", igraph::vcount(object))
    m <- paste0("Edges: ", igraph::ecount(object))
    d <- paste0("Density: ", ifelse(dens < 1e-4, "<1e-4", round(dens, 4)))
    cc <- paste0("Components: ", comps$no)
    iso <- paste0("Isolates: ", sum(igraph::degree(object) == 0))
    gstats <- paste(n, m, d, cc, iso, sep = ", ")
    gstats <- paste0(strwrap(gstats), collapse = "\n")
    gstats <- paste0(gstats, "\n")

    # graph attrs
    gattr_str <- ""
    if (length(gattrs) > 0) {
        gattr_str <- apply(cbind(names(gattrs), paste0("(", substr(sapply(gattrs, mode), 1, 1), ")")), 1, paste0, collapse = "")
        gattr_str <- paste("-Graph Attributes:\n ", paste0(gattr_str, collapse = ", "))
        gattr_str <- paste0(gattr_str, "\n", short_delim)
    }

    vattr_str <- format_attr_section(vattrs, "Vertex", short_delim)
    eattr_str <- format_attr_section(eattrs, "Edge", short_delim)
    if (igraph::ecount(object) > 0) {
        edges <- igraph::as_edgelist(object)[seq_len(min(c(10, igraph::ecount(object)))), ]
        edges <- strwrap(paste0(apply(edges, 1, paste0, collapse = c("--", "->")[igraph::is_directed(object) + 1]), collapse = " "))
        edges <- paste(edges, collapse = "\n")
        if (igraph::ecount(object) > 10) {
            edges <- c("-Edges (first 10): \n ", edges)
        } else {
            edges <- c("-Edges: \n ", edges)
        }
    } else {
        edges <- ""
    }
    cat(delim, head, delim, gstats, gattr_str, vattr_str, eattr_str, edges, sep = "")
}


format_attr_section <- function(attrs, label, short_delim) {
    if (length(attrs) == 0) return("")
    lines <- vapply(seq_along(attrs), function(l) {
        peek <- head_dot(attrs[[l]], nchar(names(attrs)[l]) + 4)
        prefix <- if (l == 1) "" else " "
        paste0(prefix, names(attrs)[l], "(", substr(mode(attrs[[l]]), 1, 1), "): ", peek, "\n")
    }, character(1))
    paste0("-", label, " Attributes:\n ", paste0(lines, collapse = ""), short_delim)
}

head_dot <- function(x, lname) {
    stri <- strwrap(paste0(x, collapse = ", "), width = 0.9 * getOption("width") - lname)[1]
    paste0(stri, " ...")
}

#' Discrete core-periphery model
#' @description Fits a discrete core-periphery model to a given network
#' @param graph igraph object
#' @param method algorithm to use (see details)
#' @param iter number of iterations if `method=GA`
#' @param ... other parameters for GA
#' @details The function fits the data to an optimal pattern matrix with a genetic algorithm (method="GA") or a rank 1 approximation, either with degree centrality (method="rk1_dc") or eigenvector centrality (method="rk1_ec") . The rank 1 approximation is computationally far cheaper but also more experimental. Best is to compare the results from both models.
#' @return list with numeric vector with entries (k1,k2,...ki...) where ki assigns vertex i to either the core (ki=1) or periphery (ki=0), and the maximal correlation with an optimal pattern matrix
#' @references
#' Borgatti, Stephen P., and Martin G. Everett. "Models of core/periphery structures." Social networks 21.4 (2000): 375-395.
#' @author David Schoch
#' @examples
#' set.seed(121)
#' # split graphs have a perfect core-periphery structure
#' sg <- split_graph(n = 20, p = 0.3, core = 0.5)
#' core_periphery(sg)
#' @export
core_periphery <- function(graph, method = "rk1_dc", iter = 500, ...) {
    A <- igraph::as_adjacency_matrix(graph, type = "both", sparse = FALSE)
    if (method == "SA") {
        warning("method='SA' is deprecated, using 'GA' instead")
        method <- "GA"
    } else if (method == "GA") {
        if (!requireNamespace("GA", quietly = TRUE)) {
            stop("The package 'GA' is needed for method='GA'")
        }

        n <- nrow(A)

        res <- GA::ga(
            "real-valued",
            fitness = cp_fct1__0,
            A = A,
            lower = rep(0, n),
            upper = rep(1, n),
            monitor = FALSE,
            maxiter = iter,
            maxFitness = 1,
            ...
        )
        return(list(
            vec = unname(round(res@solution[nrow(res@solution), ])),
            corr = res@fitnessValue
        ))
    } else if (method == "rk1_dc") {
        ev <- igraph::degree(graph, mode = "all", loops = FALSE)
        return(cp_rk1_optimize(ev, A))
    } else if (method == "rk1_ec") {
        ev <- round(igraph::eigen_centrality(graph)$vector, 8)
        return(cp_rk1_optimize(ev, A))
    } else {
        stop("method must be one of 'SA', 'rk1_dc', or 'rk1_ec'")
    }
}


# helper functions ----
cp_rk1_optimize <- function(ev, A) {
    thresh <- unique(ev)
    optcorr <- -2
    optperm <- NULL
    for (tr in thresh) {
        evabs <- (ev >= tr) + 0
        E <- outer(evabs, evabs, "+")
        E[E == 1] <- NA
        E[E == 2] <- 1
        diag(E) <- NA
        if (sum(E, na.rm = TRUE) == 0) {
            next()
        }
        tmp <- suppressWarnings(graph_cor(E, A))
        if (is.na(tmp)) {
            next()
        }
        if (tmp > optcorr) {
            optperm <- evabs
            optcorr <- tmp
        }
    }
    list(vec = optperm, corr = optcorr)
}

cp_fct1__0 <- function(A, cvec) {
    # core=1 periphery=0
    cvec <- round(cvec)
    delta <- outer(cvec, cvec, "+")
    delta[delta == 1] <- NA
    delta[delta == 2] <- 1
    diag(delta) <- NA
    suppressWarnings(graph_cor(delta, A))
}

test_that("core periphery works", {
    library(igraph)
    g <- make_full_graph(5, directed = FALSE)
    g <- add_vertices(g, 5)
    vec1 <- core_periphery(g, method = "rk1_ec")$vec
    expect_equal(vec1, c(1, 1, 1, 1, 1, 0, 0, 0, 0, 0))
    vec2 <- core_periphery(g, method = "rk1_dc")$vec
    expect_equal(vec2, c(1, 1, 1, 1, 1, 0, 0, 0, 0, 0))
    expect_error(core_periphery(graph_empty, method = "l33t"))
})

#' triad census with node attributes
#'
#' @param g igraph object. should be a directed graph
#' @param vattr name of vertex attribute to be used
#' @return triad census with node attributes
#' @details The node attribute should be integers from 1 to max(attr).
#' The output is a named vector where the names are of the form Txxx-abc, where xxx corresponds to the standard triad census notation and "abc" are the attributes of the involved nodes.
#'
#' The implemented algorithm is comparable to the algorithm in Lienert et al.
#' @references Lienert, J., Koehly, L., Reed-Tsochas, F., & Marcum, C. S. (2019). An efficient counting method for the colored triad census. Social Networks, 58, 136-142.
#' @author David Schoch
#' @examples
#' library(igraph)
#' set.seed(112)
#' g <- sample_gnp(20, p = 0.3, directed = TRUE)
#' # add a vertex attribute
#' V(g)$type <- rep(1:2, each = 10)
#' triad_census_attr(g, "type")
#' @export
triad_census_attr <- function(g, vattr) {
    if (!igraph::is_directed(g)) {
        stop("g must be a directed graph")
    }
    if (!vattr %in% igraph::vertex_attr_names(g)) {
        stop(paste0("there is no vertex attribute called ", vattr))
    }
    attr <- igraph::vertex_attr(g, vattr)
    if (!all(is.numeric(attr))) {
        stop("vertex attribute must be numeric ")
    }
    A <- igraph::as_adjacency_matrix(g)

    orbit_classes <- matrix(
        c(
            0,
            0,
            0,
            2,
            3,
            1,
            2,
            1,
            3,
            11,
            12,
            12,
            3,
            2,
            1,
            5,
            5,
            4,
            6,
            7,
            8,
            14,
            15,
            13,
            1,
            2,
            3,
            7,
            6,
            8,
            10,
            10,
            9,
            23,
            24,
            22,
            12,
            11,
            12,
            15,
            14,
            13,
            24,
            23,
            22,
            26,
            26,
            25,
            3,
            1,
            2,
            6,
            8,
            7,
            5,
            4,
            5,
            14,
            13,
            15,
            9,
            10,
            10,
            17,
            18,
            16,
            17,
            16,
            18,
            20,
            19,
            19,
            8,
            7,
            6,
            21,
            21,
            21,
            18,
            16,
            17,
            30,
            29,
            31,
            22,
            23,
            24,
            31,
            30,
            29,
            28,
            27,
            28,
            33,
            32,
            34,
            1,
            3,
            2,
            10,
            9,
            10,
            7,
            8,
            6,
            23,
            22,
            24,
            8,
            6,
            7,
            18,
            17,
            16,
            21,
            21,
            21,
            30,
            31,
            29,
            4,
            5,
            5,
            16,
            17,
            18,
            16,
            18,
            17,
            27,
            28,
            28,
            13,
            14,
            15,
            19,
            20,
            19,
            29,
            30,
            31,
            32,
            33,
            34,
            12,
            12,
            11,
            24,
            22,
            23,
            15,
            13,
            14,
            26,
            25,
            26,
            22,
            24,
            23,
            28,
            28,
            27,
            31,
            29,
            30,
            33,
            34,
            32,
            13,
            15,
            14,
            29,
            31,
            30,
            19,
            19,
            20,
            32,
            34,
            33,
            25,
            26,
            26,
            34,
            33,
            32,
            34,
            32,
            33,
            35,
            35,
            35
        ),
        ncol = 3,
        byrow = T
    )
    tri_names <- c(
        "003",
        "012",
        "012",
        "021D",
        "012",
        "102",
        "021C",
        "111U",
        "012",
        "021C",
        "021U",
        "030T",
        "021D",
        "111U",
        "030T",
        "120U",
        "012",
        "021C",
        "102",
        "111U",
        "021U",
        "111D",
        "111D",
        "201",
        "021C",
        "030C",
        "111D",
        "120C",
        "030T",
        "120C",
        "120D",
        "210",
        "012",
        "021U",
        "021C",
        "030T",
        "021C",
        "111D",
        "030C",
        "120C",
        "102",
        "111D",
        "111D",
        "120D",
        "111U",
        "201",
        "120C",
        "210",
        "021D",
        "030T",
        "111U",
        "120U",
        "030T",
        "120D",
        "120C",
        "210",
        "111U",
        "120C",
        "201",
        "210",
        "120U",
        "210",
        "210",
        "300"
    )
    tri_names <- paste0("T", tri_names)
    attrcomb <- as.matrix(expand.grid(1:max(attr), 1:max(attr), 1:max(attr)))

    code_sort <- function(vorb, attr, i) {
        idx <- order(vorb, attr)
        vorbs <- vorb[idx]
        attrs <- attr[idx]
        if (all(vorbs == 21) && length(unique(attr)) == 3) {
            if (i == 39) {
                attrs <- rev(attrs)
            }
        }
        return(paste0(
            paste0(vorbs, collapse = ""),
            "-",
            paste0(attrs, collapse = "")
        ))
    }

    triad_names <- c()
    triad_names1 <- c()
    for (i in seq_len(nrow(orbit_classes))) {
        for (j in seq_len(nrow(attrcomb))) {
            new <- code_sort(orbit_classes[i, ], attrcomb[j, ], i)
            triad_names <- unique(c(triad_names, new))
            triad_names1 <- unique(c(
                triad_names1,
                paste0(tri_names[i], "-", gsub(".*-", "", new))
            ))
        }
    }
    triads <- rep(0L, length(triad_names))
    names(triads) <- triad_names
    triads <- triadCensusCol(A, attr, orbit_classes, triads)
    names(triads) <- triad_names1
    triads
}

test_that("triad census attr works", {
    library(igraph)
    gundir <- make_full_graph(3, directed = FALSE)
    gnoatt <- make_full_graph(3, directed = TRUE)
    gchatt <- make_full_graph(3, directed = TRUE)
    vertex_attr(gchatt, "test") <- c("a", "b", "c")
    gtriad <- make_full_graph(3, directed = TRUE)
    vertex_attr(gtriad, "test") <- 1:3
    expect_error(triad_census_attr(gundir))
    expect_error(triad_census_attr(gnoatt))
    expect_error(triad_census_attr(gchatt, "test"))
    expect_true(triad_census_attr(gtriad, "test")["T300-123"] == 1)
})

#' triad census with node attributes
#'
#' @param g igraph object. should be a directed graph
#' @param vattr name of vertex attribute to be used
#' @return triad census with node attributes
#' @details The node attribute should be integers from 1 to max(attr).
#' The output is a named vector where the names are of the form Txxx-abc, where xxx corresponds to the standard triad census notation and "abc" are the attributes of the involved nodes.
#'
#' The implemented algorithm is comparable to the algorithm in Lienert et al.
#' @references Lienert, J., Koehly, L., Reed-Tsochas, F., & Marcum, C. S. (2019). An efficient counting method for the colored triad census. Social Networks, 58, 136-142.
#' @author David Schoch
#' @examples
#' library(igraph)
#' set.seed(112)
#' g <- sample_gnp(20, p = 0.3, directed = TRUE)
#' # add a vertex attribute
#' V(g)$type <- rep(1:2, each = 10)
#' triad_census_attr(g, "type")
#' @export
triad_census_attr <- function(g, vattr) {
    attr <- validate_vattr(g, vattr)
    A <- igraph::as_adjacency_matrix(g)

    # orbit codes (sorted triple) for each of the 64 directed triad realizations
    orbit_classes <- matrix(
        c(
             0,  0,  0,   2,  3,  1,   2,  1,  3,  11, 12, 12,
             3,  2,  1,   5,  5,  4,   6,  7,  8,  14, 15, 13,
             1,  2,  3,   7,  6,  8,  10, 10,  9,  23, 24, 22,
            12, 11, 12,  15, 14, 13,  24, 23, 22,  26, 26, 25,
             3,  1,  2,   6,  8,  7,   5,  4,  5,  14, 13, 15,
             9, 10, 10,  17, 18, 16,  17, 16, 18,  20, 19, 19,
             8,  7,  6,  21, 21, 21,  18, 16, 17,  30, 29, 31,
            22, 23, 24,  31, 30, 29,  28, 27, 28,  33, 32, 34,
             1,  3,  2,  10,  9, 10,   7,  8,  6,  23, 22, 24,
             8,  6,  7,  18, 17, 16,  21, 21, 21,  30, 31, 29,
             4,  5,  5,  16, 17, 18,  16, 18, 17,  27, 28, 28,
            13, 14, 15,  19, 20, 19,  29, 30, 31,  32, 33, 34,
            12, 12, 11,  24, 22, 23,  15, 13, 14,  26, 25, 26,
            22, 24, 23,  28, 28, 27,  31, 29, 30,  33, 34, 32,
            13, 15, 14,  29, 31, 30,  19, 19, 20,  32, 34, 33,
            25, 26, 26,  34, 33, 32,  34, 32, 33,  35, 35, 35
        ),
        ncol = 3,
        byrow = TRUE
    )
    tri_names <- c(
        "003", "012", "012", "021D", "012", "102", "021C", "111U",
        "012", "021C", "021U", "030T", "021D", "111U", "030T", "120U",
        "012", "021C", "102", "111U", "021U", "111D", "111D", "201",
        "021C", "030C", "111D", "120C", "030T", "120C", "120D", "210",
        "012", "021U", "021C", "030T", "021C", "111D", "030C", "120C",
        "102", "111D", "111D", "120D", "111U", "201", "120C", "210",
        "021D", "030T", "111U", "120U", "030T", "120D", "120C", "210",
        "111U", "120C", "201", "210", "120U", "210", "210", "300"
    )
    tri_names <- paste0("T", tri_names)
    attrcomb <- as.matrix(expand.grid(1:max(attr), 1:max(attr), 1:max(attr)))

    code_sort <- function(vorb, attr, i) {
        idx <- order(vorb, attr)
        vorbs <- vorb[idx]
        attrs <- attr[idx]
        if (all(vorbs == 21) && length(unique(attr)) == 3) {
            if (i == 39) {
                attrs <- rev(attrs)
            }
        }
        return(paste0(
            paste0(vorbs, collapse = ""),
            "-",
            paste0(attrs, collapse = "")
        ))
    }

    # Build all (orbit class, attribute combination) names into preallocated
    # vectors and deduplicate once. The "new" code (sorted orbits + attrs) maps
    # bijectively to its labelled name, so a single first-occurrence mask keeps
    # both vectors in lockstep.
    n_comb <- nrow(orbit_classes) * nrow(attrcomb)
    all_codes <- character(n_comb)
    all_labels <- character(n_comb)
    idx <- 1L
    for (i in seq_len(nrow(orbit_classes))) {
        for (j in seq_len(nrow(attrcomb))) {
            new <- code_sort(orbit_classes[i, ], attrcomb[j, ], i)
            all_codes[idx] <- new
            all_labels[idx] <- paste0(tri_names[i], "-", gsub(".*-", "", new))
            idx <- idx + 1L
        }
    }
    keep <- !duplicated(all_codes)
    triad_names <- all_codes[keep]
    triad_names1 <- all_labels[keep]

    triads <- rep(0L, length(triad_names))
    names(triads) <- triad_names
    triads <- triadCensusCol(A, attr, orbit_classes, triads)
    names(triads) <- triad_names1
    triads
}

test_that("str.igraph prints without error and includes key sections", {
    library(igraph)
    set.seed(1)
    g <- sample_gnp(8, 0.4, directed = TRUE)
    g$name <- "test net"
    V(g)$grp <- rep(1:2, 4)
    E(g)$weight <- seq_len(ecount(g))
    out <- paste(capture.output(str(g)), collapse = "\n")
    expect_true(grepl("TEST NET", out))
    expect_true(grepl("directed", out))
    expect_true(grepl("weighted", out))
    expect_true(grepl("Nodes: 8", out))
    expect_true(grepl("Graph Attributes", out))
    expect_true(grepl("Vertex Attributes", out))
    expect_true(grepl("Edge Attributes", out))
})

test_that("str.igraph handles a minimal graph with no attributes or edges", {
    library(igraph)
    g <- make_empty_graph(n = 3, directed = FALSE)
    out <- paste(capture.output(str(g)), collapse = "\n")
    expect_true(grepl("UNNAMED NETWORK", out))
    expect_true(grepl("Edges: 0", out))
    expect_true(grepl("Isolates: 3", out))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast_cliques.R
\name{fast_cliques}
\alias{fast_cliques}
\title{Find Cliques, maximal or not, fast}
\usage{
fast_cliques(g, what = "M", min = NULL, max = NULL, outfile = NA)
}
\arguments{
\item{g}{An igraph object}

\item{what}{either "M" for maximal cliques or "C" for all cliques}

\item{min}{Numeric constant, lower limit on the size of the cliques to find. NULL means no limit, ie. it is the same as 0}

\item{max}{Numeric constant, upper limit on the size of the cliques to find. NULL means no limit}

\item{outfile}{character. If not NA, cliques are written to file}
}
\value{
a list containing numeric vectors of vertex ids. Each list element is a clique. If outfile!=NA, the output is written to the specified file
}
\description{
Enumerates all (maximal) cliques using MACE. Can be faster than igraph in some circumstances
}
\details{
C Code downloaded from http://research.nii.ac.jp/~uno/codes.htm. Download the code and run make and then point an environment variable called MACE_PATH to the binary. See http://research.nii.ac.jp/~uno/code/mace.html for more details. MACE is faster than igraph for dense graphs.
}
\references{
Kazuhisa Makino, Takeaki Uno, "New Algorithms for Enumerating All Maximal Cliques", Lecture Notes in Computer Science 3111 (Proceedings of SWAT 2004), Springer, pp.260-272, 2004
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/triad_census_attr.R
\name{triad_census_attr}
\alias{triad_census_attr}
\title{triad census with node attributes}
\usage{
triad_census_attr(g, vattr)
}
\arguments{
\item{g}{igraph object. should be a directed graph}

\item{vattr}{name of vertex attribute to be used}
}
\value{
triad census with node attributes
}
\description{
triad census with node attributes
}
\details{
The node attribute should be integers from 1 to max(attr).
The output is a named vector where the names are of the form Txxx-abc, where xxx corresponds to the standard triad census notation and "abc" are the attributes of the involved nodes.

The implemented algorithm is comparable to the algorithm in Lienert et al.
}
\examples{
library(igraph)
set.seed(112)
g <- sample_gnp(20, p = 0.3, directed = TRUE)
# add a vertex attribute
V(g)$type <- rep(1:2, each = 10)
triad_census_attr(g, "type")
}
\references{
Lienert, J., Koehly, L., Reed-Tsochas, F., & Marcum, C. S. (2019). An efficient counting method for the colored triad census. Social Networks, 58, 136-142.
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_structures.R
\name{clique_vertex_mat}
\alias{clique_vertex_mat}
\title{Clique Vertex Matrix}
\usage{
clique_vertex_mat(g)
}
\arguments{
\item{g}{An igraph object}
}
\value{
Numeric matrix
}
\description{
Creates the clique vertex matrix with entries (i,j) equal to one if node j is in clique i
}
\examples{
library(igraph)
g <- sample_gnp(10, 0.2)
clique_vertex_mat(g)
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lfr_benchmark.R
\name{sample_lfr}
\alias{sample_lfr}
\title{LFR benchmark graphs}
\usage{
sample_lfr(
  n,
  tau1 = 2,
  tau2 = 1,
  mu = 0.1,
  average_degree,
  max_degree,
  min_community = NULL,
  max_community = NULL,
  on = 0,
  om = 0
)
}
\arguments{
\item{n}{Number of nodes in the created graph.}

\item{tau1}{Power law exponent for the degree distribution of the created graph. This value must be strictly greater than one}

\item{tau2}{Power law exponent for the community size distribution in the created graph. This value must be strictly greater than one}

\item{mu}{Fraction of inter-community edges incident to each node. This value must be in the interval 0 to 1.}

\item{average_degree}{Desired average degree of nodes in the created graph. This value must be in the interval 0 to n. Exactly one of this and \code{min_degree} must be specified, otherwise an error is raised}

\item{max_degree}{Maximum degree of nodes in the created graph. If not specified, this is set to n-1.}

\item{min_community}{Minimum size of communities in the graph. If not specified, this is set to \code{min_degree}}

\item{max_community}{Maximum size of communities in the graph. If not specified, this is set to n, the total number of nodes in the graph.}

\item{on}{number of overlapping nodes}

\item{om}{number of memberships of the overlapping nodes}
}
\value{
an igraph object
}
\description{
Generates benchmark networks for clustering tasks with a priori known communities. The algorithm accounts for the heterogeneity in the distributions of node degrees and of community sizes.
}
\details{
code adapted from \url{https://github.com/synwalk/synwalk-analysis/tree/master/lfr_generator}
}
\examples{
# Simple Girven-Newman benchmark graphs
g <- sample_lfr(
    n = 128, average_degree = 16,
    max_degree = 16, mu = 0.1,
    min_community = 32, max_community = 32
)
}
\references{
A. Lancichinetti, S. Fortunato, and F. Radicchi.(2008) Benchmark graphs for testing community detection algorithms. Physical Review E, 78. arXiv:0805.4770
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_products.R
\name{graph_direct}
\alias{graph_direct}
\title{Direct product of two graphs}
\usage{
graph_direct(g, h)
}
\arguments{
\item{g}{An igraph object}

\item{h}{An igraph object}
}
\value{
Direct product as igraph object
}
\description{
Compute the direct product of two graphs
}
\details{
See https://en.wikipedia.org/wiki/Tensor_product_of_graphs
}
\examples{
library(igraph)
g <- make_ring(4)
h <- make_full_graph(2)
graph_direct(g, h)
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structural_equivalence.R
\name{structural_equivalence}
\alias{structural_equivalence}
\title{Maximal Structural Equivalence}
\usage{
structural_equivalence(g)
}
\arguments{
\item{g}{An igraph object}
}
\value{
vector of equivalence classes
}
\description{
Calculates structural equivalence for an undirected graph
}
\details{
Two nodes u and v are structurally equivalent if they have exactly the same neighbors. The equivalence classes produced with this function are either cliques or empty graphs.
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_pa_homophilic.R
\name{sample_pa_homophilic}
\alias{sample_pa_homophilic}
\title{Homophilic random graph using BA preferential attachment model}
\usage{
sample_pa_homophilic(
  n,
  m,
  minority_fraction,
  h_ab,
  h_ba = NULL,
  directed = FALSE
)
}
\arguments{
\item{n}{number of nodes}

\item{m}{number of edges a new node is connected to}

\item{minority_fraction}{fraction of nodes that belong to the minority group}

\item{h_ab}{probability to connect a node from group a with groub b}

\item{h_ba}{probability to connect a node from group b with groub a. If NULL, h_ab is used.}

\item{directed}{should a directed network be created}
}
\value{
igraph object
}
\description{
A graph of n nodes is grown by attaching new nodes each with m
edges that are preferentially attached to existing nodes with high
degree, depending on the homophily parameters.
}
\details{
The code is an adaption of the python code from https://github.com/gesiscss/HomophilicNtwMinorities/
}
\references{
Karimi, F., Génois, M., Wagner, C., Singer, P., & Strohmaier, M. (2018). Homophily influences ranking of minorities in social networks. Scientific reports, 8(1), 1-12. (https://www.nature.com/articles/s41598-018-29405-7)

Espín-Noboa, L., Wagner, C., Strohmaier, M., & Karimi, F. (2022). Inequality and inequity in network-based ranking and recommendation algorithms. Scientific reports, 12(1), 1-14. (https://www.nature.com/articles/s41598-022-05434-1)
}
\author{
David Schoch
#maximally heterophilic network
sample_pa_homophilic(n = 50, m = 2,minority_fraction = 0.2,h_ab = 1)
#maximally homophilic network
sample_pa_homophilic(n = 50, m = 2,minority_fraction = 0.2,h_ab = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_structures.R
\name{as_adj_list1}
\alias{as_adj_list1}
\title{Adjacency list}
\usage{
as_adj_list1(g)
}
\arguments{
\item{g}{An igraph object}
}
\value{
A list of numeric vectors.
}
\description{
Create adjacency lists from a graph, either for adjacent edges or for neighboring vertices. This version is faster than the version of igraph but less general.
}
\details{
The function does not have a mode parameter and only returns the adjacency list comparable to as_adj_list(g,mode="all)
}
\examples{
library(igraph)
g <- make_ring(10)
as_adj_list1(g)
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{helpers}
\alias{helpers}
\alias{biggest_component}
\alias{delete_isolates}
\title{helper function}
\usage{
biggest_component(g)

delete_isolates(g)
}
\arguments{
\item{g}{igraph object}
}
\value{
igraph object
}
\description{
small functions to deal with typical network problems
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_kcores.R
\name{sample_coreseq}
\alias{sample_coreseq}
\title{Generate random graphs with a given coreness sequence}
\usage{
sample_coreseq(cores)
}
\arguments{
\item{cores}{coreness sequence}
}
\value{
igraph object of graph with the same coreness sequence as the input
}
\description{
Similar to \link[igraph]{sample_degseq} just with \link[igraph]{coreness}
}
\details{
The code is an adaption of the python code from https://github.com/ktvank/Random-Graphs-with-Prescribed-K-Core-Sequences/
}
\examples{
library(igraph)
g1 <- make_graph("Zachary")
kcores1 <- coreness(g1)
g2 <- sample_coreseq(kcores1)
kcores2 <- coreness(g2)

# the sorted arrays are the same
all(sort(kcores1) == sort(kcores2))

}
\references{
Van Koevering, Katherine, Austin R. Benson, and Jon Kleinberg. 2021. ‘Random Graphs with Prescribed K-Core Sequences: A New Null Model for Network Analysis’. ArXiv:2102.12604. https://doi.org/10.1145/3442381.3450001.
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphs.R
\name{bipartite_from_data_frame}
\alias{bipartite_from_data_frame}
\title{two-mode network from a data.frame}
\usage{
bipartite_from_data_frame(d, type1, type2, attr = NULL, weighted = TRUE)
}
\arguments{
\item{d}{data.frame}

\item{type1}{column name of mode 1}

\item{type2}{column name of mode 2}

\item{attr}{named list of edge attributes}

\item{weighted}{should a weighted graph be created if multiple edges occur}
}
\value{
two mode network as igraph object
}
\description{
Create a two-mode network from a data.frame
}
\examples{
library(igraph)
edges <- data.frame(mode1 = 1:5, mode2 = letters[1:5])
bipartite_from_data_frame(edges, "mode1", "mode2")
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphs.R
\name{graph_kpartite}
\alias{graph_kpartite}
\title{k partite graphs}
\usage{
graph_kpartite(n = 10, grp = c(5, 5))
}
\arguments{
\item{n}{number of nodes}

\item{grp}{vector of partition sizes}
}
\value{
igraph object
}
\description{
Create a random k-partite graph.
}
\examples{
# 3-partite graph with equal sized groups
graph_kpartite(n = 15, grp = c(5, 5, 5))
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dyad_census_attr.R
\name{dyad_census_attr}
\alias{dyad_census_attr}
\title{dyad census with node attributes}
\usage{
dyad_census_attr(g, vattr)
}
\arguments{
\item{g}{igraph object. should be a directed graph.}

\item{vattr}{name of vertex attribute to be used.}
}
\value{
dyad census as a data.frame.
}
\description{
dyad census with node attributes
}
\details{
The node attribute should be integers from 1 to max(attr)
}
\examples{
library(igraph)
g <- sample_gnp(10, 0.4, directed = TRUE)
V(g)$attr <- c(rep(1, 5), rep(2, 5))
dyad_census_attr(g, "attr")
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_structures.R
\name{as_adj_weighted}
\alias{as_adj_weighted}
\title{weighted dense adjacency matrix}
\usage{
as_adj_weighted(g, attr = NULL)
}
\arguments{
\item{g}{An igraph object}

\item{attr}{Either NULL or a character string giving an edge attribute name. If NULL a traditional adjacency matrix is returned. If not NULL then the values of the given edge attribute are included in the adjacency matrix.}
}
\value{
Numeric matrix
}
\description{
returns the weighted adjacency matrix in dense format
}
\details{
This method is faster than as_adj from igraph if you need the weighted adjacency matrix in dense format
}
\examples{
library(igraph)
g <- sample_gnp(10, 0.2)
E(g)$weight <- runif(ecount(g))
as_adj_weighted(g, attr = "weight")
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_igraph.R
\name{str.igraph}
\alias{str.igraph}
\title{Print graphs to terminal}
\usage{
\method{str}{igraph}(object, ...)
}
\arguments{
\item{object}{An igraph object}

\item{...}{additional arguments to print (ignored)}
}
\value{
str does not return anything. The obvious side effect is output to the terminal.
}
\description{
Prints an igraph object to terminal (different than the standard igraph method)
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reciprocity.R
\name{reciprocity_cor}
\alias{reciprocity_cor}
\title{Reciprocity correlation coefficient}
\usage{
reciprocity_cor(g)
}
\arguments{
\item{g}{igraph object. should be a directed graph}
}
\value{
Reciprocity as a correlation
}
\description{
Reciprocity correlation coefficient
}
\details{
The usual definition of reciprocity has some defects. It cannot tell the relative difference of reciprocity compared with purely random network with the same number of vertices and edges.
The useful information from reciprocity is not the value itself, but whether mutual links occur more or less often than expected by chance.

To overcome this issue, reciprocity can be defined as the correlation coefficient between the entries of the adjacency matrix of a directed graph:
\deqn{
 \frac{\sum_{i\neq j} (a_{ij} - a')((a_{ji} - a')}{\sum_{i\neq j} (a_{ij} - a')^2}
}
where a' is the density of g.

This definition gives an absolute quantity which directly allows one to distinguish between reciprocal (>0) and antireciprocal (< 0) networks, with mutual links occurring more and less often than random respectively.
}
\examples{
library(igraph)
g <- sample_gnp(20, p = 0.3, directed = TRUE)
reciprocity(g)
reciprocity_cor(g)
}
\references{
Diego Garlaschelli; Loffredo, Maria I. (2004). "Patterns of Link Reciprocity in Directed Networks". Physical Review Letters. American Physical Society. 93 (26): 268701
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphs.R
\name{split_graph}
\alias{split_graph}
\title{split graph}
\usage{
split_graph(n, p, core)
}
\arguments{
\item{n}{number of nodes}

\item{p}{probability of peripheral nodes to connect to the core nodes}

\item{core}{fraction of nodes in the core}
}
\value{
igraph object
}
\description{
Create a random split graph with a perfect core-periphery structure.
}
\examples{
# split graph with 20 nodes and a core size of 10
split_graph(n = 20, p = 0.4, 0.5)
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qap.R
\name{graph_cor}
\alias{graph_cor}
\alias{graph_cor.default}
\alias{graph_cor.igraph}
\alias{graph_cor.matrix}
\alias{graph_cor.array}
\title{Graph correlation}
\usage{
graph_cor(object1, object2)

\method{graph_cor}{default}(object1, object2)

\method{graph_cor}{igraph}(object1, object2, ...)

\method{graph_cor}{matrix}(object1, object2)

\method{graph_cor}{array}(object1, object2)
}
\arguments{
\item{object1}{igraph object or adjacency matrix}

\item{object2}{igraph object or adjacency matrix over the same vertex set as object1}

\item{...}{additional arguments}
}
\value{
correlation between graphs
}
\description{
This function computes the correlation between networks. Implemented methods expect
the graph to be an adjacency matrix, an igraph, or a network object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core_periphery.R
\name{core_periphery}
\alias{core_periphery}
\title{Discrete core-periphery model}
\usage{
core_periphery(graph, method = "rk1_dc", iter = 500, ...)
}
\arguments{
\item{graph}{igraph object}

\item{method}{algorithm to use (see details)}

\item{iter}{number of iterations if \code{method=GA}}

\item{...}{other parameters for GA}
}
\value{
list with numeric vector with entries (k1,k2,...ki...) where ki assigns vertex i to either the core (ki=1) or periphery (ki=0), and the maximal correlation with an optimal pattern matrix
}
\description{
Fits a discrete core-periphery model to a given network
}
\details{
The function fits the data to an optimal pattern matrix with a genetic algorithm (method="GA") or a rank 1 approximation, either with degree centrality (method="rk1_dc") or eigenvector centrality (method="rk1_ec") . The rank 1 approximation is computationally far cheaper but also more experimental. Best is to compare the results from both models.
}
\examples{
set.seed(121)
# split graphs have a perfect core-periphery structure
sg <- split_graph(n = 20, p = 0.3, core = 0.5)
core_periphery(sg)
}
\references{
Borgatti, Stephen P., and Martin G. Everett. "Models of core/periphery structures." Social networks 21.4 (2000): 375-395.
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphs.R
\name{graph_from_multi_edgelist}
\alias{graph_from_multi_edgelist}
\title{Multiple networks from a single edgelist with a typed attribute}
\usage{
graph_from_multi_edgelist(
  d,
  from = NULL,
  to = NULL,
  type = NULL,
  weight = NULL,
  directed = FALSE
)
}
\arguments{
\item{d}{data frame.}

\item{from}{column name of sender. If NULL, defaults to first column.}

\item{to}{column of receiver. If NULL, defaults to second column.}

\item{type}{type attribute to split the edgelist. If NULL, defaults to third column.}

\item{weight}{optional column name of edge weights. Ignored if NULL.}

\item{directed}{logical scalar, whether or not to create a directed graph.}
}
\value{
list of igraph objects.
}
\description{
Create a list of igraph objects from an edgelist according to a type attribute
}
\examples{
library(igraph)
d <- data.frame(
    from = rep(c(1, 2, 3), 3), to = rep(c(2, 3, 1), 3),
    type = rep(c("a", "b", "c"), each = 3), weight = 1:9
)
graph_from_multi_edgelist(d, "from", "to", "type", "weight")
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_products.R
\name{graph_cartesian}
\alias{graph_cartesian}
\title{Cartesian product of two graphs}
\usage{
graph_cartesian(g, h)
}
\arguments{
\item{g}{An igraph object}

\item{h}{An igraph object}
}
\value{
Cartesian product as igraph object
}
\description{
Compute the Cartesian product of two graphs
}
\details{
See https://en.wikipedia.org/wiki/Cartesian_product_of_graphs
}
\examples{
library(igraph)
g <- make_ring(4)
h <- make_full_graph(2)
graph_cartesian(g, h)
}
\author{
David Schoch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_structures.R
\name{as_multi_adj}
\alias{as_multi_adj}
\title{Convert a list of graphs to an adjacency matrices}
\usage{
as_multi_adj(g_lst, attr = NULL, sparse = FALSE)
}
\arguments{
\item{g_lst}{A list of igraph object}

\item{attr}{Either NULL or a character string giving an edge attribute name. If NULL a binary adjacency matrix is returned.}

\item{sparse}{Logical scalar, whether to create a sparse matrix. The 'Matrix' package must be installed for creating sparse matrices.}
}
\value{
List of numeric matrices
}
\description{
Convenience function that turns a list of igraph objects into adjacency matrices.
}
\author{
David Schoch
}
